#coding: UTF-8
## Copyright 2009-2011 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

from cgi import escape

from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.utils.translation import ugettext as _
from django.conf import settings

import lino

from lino import reports, actions, fields
from lino.utils import constrain
from lino.utils import jsgen
from lino.utils.jsgen import py2js, Variable, Component, id2js, js_code
from lino.utils import choosers
#~ from . import ext_requests
from lino.ui import requests as ext_requests

EXT_CHAR_WIDTH = 9
EXT_CHAR_HEIGHT = 22

#~ DEFAULT_GC_NAME = 'std'
DEFAULT_GC_NAME = 0


def a2btn(a):
    return dict(
      opens_a_slave=a.opens_a_slave,
      handler=js_code("Lino.%s" % a),
      name=a.name,
      label=unicode(a.label),
      #~ url="/".join(("/ui",a.actor.app_label,a.actor._actor_name,a.name))
    )
      
def py2html(obj,name):
    for n in name.split('.'):
        obj = getattr(obj,n,"N/A")
    return escape(unicode(obj))
    
def before_row_edit(panel):
    l = []
    #~ l.append("console.log('before_row_edit',record);")
    for e in panel.active_children:
        if isinstance(e,GridElement):
            l.append("%s.on_master_changed();" % e.as_ext())
        elif isinstance(e,PictureElement):
            l.append("this.load_picture_to(%s,record);" % e.as_ext())
        elif isinstance(e,HtmlBoxElement):
            l.append("%s.refresh();" % e.as_ext())
        elif isinstance(e,FieldElement):
            chooser = choosers.get_for_field(e.field)
            if chooser:
                #~ logger.debug("20100615 %s.%s has chooser", self.lh.layout, e.field.name)
                for f in chooser.context_fields:
                    #~ l.append("console.log('20110128 before_row_edit',record.data);")
                    l.append("%s.setContextValue(%r,record ? record.data[%r] : undefined);" % (
                        e.ext_name,f.name,ext_requests.form_field_name(f)))
    #~ return js_code('function(record){\n  %s\n}' % ('\n  '.join(l)))
    return js_code('function(record){ %s }' % (' '.join(l)))

class GridColumn(Component):
    declare_type = jsgen.DECLARE_INLINE
    field = None
    
    def __init__(self,index,editor,**kw):
        """editor may be a Panel for columns on a GenericForeignKey
        """
        #~ print 20100515, editor.name, editor.__class__
        #~ assert isinstance(editor,FieldElement), \
            #~ "%s.%s is a %r (expected FieldElement instance)" % (cm.grid.report,editor.name,editor)
        self.editor = editor
        #~ self.value_template = editor.grid_column_template
        kw.update(sortable=True)
        self.index = index
        self.label = editor.label
        kw.update(self.editor.get_column_options())
        kw.update(hidden=editor.hidden)
        if settings.USE_GRIDFILTERS and editor.filter_type:
            kw.update(filter=dict(type=editor.filter_type))
        #~ if isinstance(editor,FieldElement) and editor.field.primary_key:
        if isinstance(editor,FieldElement):
            self.field = editor.field
            rend = None
            if isinstance(editor.field,models.AutoField):
                rend = 'Lino.id_renderer'
                #~ kw.update(renderer=js_code('Lino.id_renderer'))
            elif isinstance(editor.field,models.ForeignKey):
                # FK fields are clickable if their target has a detail view
                rpt = editor.field.rel.to._lino_model_report
                if rpt.detail_action is not None:
                #~ a = rpt.get_action('detail')
                #~ if a is not None:
                    rend = "Lino.fk_renderer('%s','Lino.%s')" % (
                      editor.field.name + 'Hidden',
                      rpt.detail_action)
                    #~ rend = "Lino.fk_renderer('%s','%s')" % (
                      #~ editor.field.name + 'Hidden',
                      #~ editor.lh.rh.ui.get_actor_url(rpt))
            #~ if not rend:
                #~ rend = 'Lino.default_renderer'
            if rend:
                kw.update(renderer=js_code(rend))
            kw.update(editable=editor.editable)
            if editor.editable:
                kw.update(editor=editor)
        else:
            kw.update(editable=False)
        Component.__init__(self,editor.name,**kw)
    
        #~ if self.editable:
            #~ editor = self.get_field_options()
            
        
class Toolbar(Component):
    value_template = "new Ext.Toolbar(%s)"
class ComboBox(Component):
    value_template = 'new Ext.form.ComboBox(%s)'
class ExtPanel(Component): # todo: rename this to Panel, and Panel to PanelElement or sth else
    value_template = "new Ext.Panel(%s)"
    
                    
        
        
class VisibleComponent(Component):
    vflex = False
    hflex = True
    width = None
    height = None
    preferred_width = 10
    preferred_height = 1
    #flex = None
    
    def __init__(self,name,width=None,height=None,label=None,**kw):
        Component.__init__(self,name,**kw)
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height
        if label is not None:
            self.label = label
        #~ if flex is not None:
            #~ self.flex = flex
    

    def __str__(self):
        "This shows how elements are specified"
        name = Component.__str__(self)
        if self.width is None:
            return name
        if self.height is None:
            return name + ":%d" % self.width
        return name + ":%dx%d" % (self.width,self.height)
        
    def __repr__(self):
        return str(self)
        
    def pprint(self,level=0):
        return ("  " * level) + str(self)
        
    def walk(self):
        yield self
        
        
    def debug_lines(self):
        sep = u"</td><td>"
        cols = "ext_name name parent label __class__.__name__ labelAlign vertical width preferred_width height preferred_height flex".split()
        yield '<tr><td>' + sep.join(cols) + '</td></tr>'
        for e in self.walk():
            yield '<tr><td>'+sep.join([py2html(e,n) for n in cols]) +'</td></tr>'
            
    def has_field(self,fld):
        for de in self.walk():
            if isinstance(de,FieldElement) and de.field is fld:
                return True
        return False
        
        
class LayoutElement(VisibleComponent):
    stored = False
    ext_name = None
    ext_suffix = ""
    data_type = None 
    filter_type = None
    parent = None # will be set by Container
    field = None
    
    label = None
    label_width = 0 
    editable = False
    sortable = False
    xtype = None # set by subclasses
    #~ grid_column_template = "new Ext.grid.Column(%s)"
    collapsible = False
    hidden = False
    active_child = True
    refers_to_ww = False
    
    def __init__(self,lh,name,**kw):
        #logger.debug("LayoutElement.__init__(%r,%r)", lh.layout,name)
        #self.parent = parent
        #~ name = lh.layout._actor_name + '_' + name
        VisibleComponent.__init__(self,name,**kw)
        self.lh = lh
        #~ if lh is not None:
        assert isinstance(lh,reports.LayoutHandle)
        #~ lh.setup_element(self)

    def submit_fields(self):
        return []
        
    def update_config(self,wc):
        pass
        
    def get_property(self,name):
        v = getattr(self,name,None)
        if self.parent is None or v is not None:
            return v
        return self.parent.get_property(name)
        
    def get_column_options(self,**kw):
        return kw
        
    def set_parent(self,parent):
        #~ if self.parent is not None:
            #~ raise Exception("%s : parent is already %s, cannot set it to %s" % (self,self.parent,parent))
        self.parent = parent
        if self.label and isinstance(parent,Panel):
            if parent.labelAlign == reports.LABEL_ALIGN_LEFT:
                self.preferred_width += len(self.label)

    def ext_options(self,**kw):
        kw = VisibleComponent.ext_options(self,**kw)
        if self.xtype is not None:
            kw.update(xtype=self.xtype)
        if self.collapsible:
            kw.update(collapsible=True)
        return kw


class DataViewElement(LayoutElement):
    declare_type = jsgen.DECLARE_INLINE
    value_template = "new Ext.DataView(%s)"
    vflex = True
    
    def __init__(self,lh,name,dv,**kw):
        #~ self.dv = dv
        kw.update(store=js_code('this.store'))
        kw.update(tpl=js_code('new Ext.XTemplate(%s)' % py2js(dv.xtemplate)))
        kw.update(emptyText='No images to display')
        kw.update(itemSelector='div.thumb-wrap')
        kw.update(loadingText='Loading...')
        LayoutElement.__init__(self,lh,name,**kw)

class TemplateElement(LayoutElement):
    declare_type = jsgen.DECLARE_INLINE
    value_template = "new Ext.BoxComponent(%s)"
    vflex = True
    
    def __init__(self,lh,name,dv,**kw):
        #~ self.dv = dv
        #~ kw.update(tpl=js_code('new Ext.XTemplate(%s)' % py2js(dv.xtemplate)))
        kw.update(plugins=js_code('new Lino.TemplateBoxPlugin(caller,%s)' % py2js(dv.xtemplate)))
        LayoutElement.__init__(self,lh,name,**kw)

class PictureElement(LayoutElement):
    declare_type = jsgen.DECLARE_VAR
    #~ declare_type = jsgen.DECLARE_INLINE
    value_template = "new Ext.BoxComponent(%s)"
    vflex = False
    hflex = False
    #~ vflex = True
    
    def __init__(self,lh,name,action,**kw):
        #~ print 20100730, name
        #~ kw.update(html='<img height="100%"/>')
        #~ kw.update(html='<img height="100%" onclick="Lino.img_onclick()"/>')
        #~ kw.update(autoHeight=True)
        #~ kw.update(anchor="100% 100%")
        kw.update(autoEl=dict(tag='img'))
        #~ kw.update(autoEl=dict(tag='img',onclick="Lino.img_onclick(%s)" % name))
        #~ kw.update(autoEl=dict(tag='img',onclick="Lino.img_onclick(this)" ))
        #~ kw.update(onclick=js_code('"Lino.img_onclick()"'))
        #~ kw.update(cls='ext-el-mask')
        #~ kw.update(style=dict(width='100px',cursor='pointer'))
        #~ kw.update(style=dict(width='100px',height='150px',cursor='pointer'))
        kw.update(plugins=js_code('Lino.PictureBoxPlugin'))
        #~ kw.update(plugins=js_code('new Lino.PictureBoxPlugin(caller)'))
        #~ kw.update(listeners=dict(click=js_code('Lino.img_onclick')))
        LayoutElement.__init__(self,lh,name,**kw)
        #~ print "PictureElement", self.width, self.preferred_width
        self.update(style=dict(
          width='%dpx' % ((self.width or self.preferred_width)*10), # assume 1 char ~ 10px
          cursor='pointer'))
        

class unused_ButtonElement(LayoutElement):
    value_template = "new Ext.Component(%s)"
    
    def __init__(self,lh,name,meth,label,**kw):
        self.label = label
        self.meth = meth
        #~ onclick='alert("oops")'
        #~ kw.update(html='<input type="button" onclick="%s" value=" %s ">' % (onclick,label))
        LayoutElement.__init__(self,lh,name,**kw)
    
    def get_column_options(self,**kw):
        #~ kw.update(dataIndex=self.field.name)
        #~ if self.label is None:
            #~ kw.update(header=self.field.name)
        #~ else:
        kw.update(header=unicode(self.label))
        kw.update(editable=False)
        kw.update(sortable=False)
        rend = 'Lino.cell_button_renderer'
        kw.update(renderer=js_code(rend))
        return kw    
                
        

class StaticTextElement(LayoutElement):
    declare_type = jsgen.DECLARE_INLINE
    #declare_type = jsgen.DECLARE_THIS
    #~ declare_type = jsgen.DECLARE_VAR
    xtype = 'label'
    
    def __init__(self,lh,name,text,**kw):
        kw.update(html=text.text)
        LayoutElement.__init__(self,lh,name,**kw)
        #~ self.text = text

    #~ def ext_options(self,**kw):
        #~ kw = LayoutElement.ext_options(self,**kw)
        #~ kw.update(html=self.text.text)
        #~ return kw
        
        
        

class Spacer(LayoutElement):
    declare_type = jsgen.DECLARE_INLINE
    #xtype = 'label'
    value_template = "new Ext.Spacer(%s)"
    
        
        
        
class FieldElement(LayoutElement):
    #~ declare_type = jsgen.DECLARE_INLINE
    #~ declare_type = jsgen.DECLARE_THIS
    declare_type = jsgen.DECLARE_VAR
    stored = True
    filter_type = None # 'auto'
    #declaration_order = 3
    #~ ext_suffix = "_field"
    
    def __init__(self,lh,field,**kw):
        assert field.name, Exception("field %r has no name!" % field)
        self.field = field
        self.editable = field.editable # and not field.primary_key
        #~ LayoutElement.__init__(self,lh,varname_field(field),label=unicode(field.verbose_name),**kw)
        LayoutElement.__init__(self,lh,field.name,label=unicode(field.verbose_name),**kw)
        
    #~ def get_filter_options(self,**kw):
        #~ if self.filter_type:
            #~ kw.update(dataIndex=self.field.name)
            # 20100805 see also GridFilters.js
            #~ kw.update(filterable=True)  
            #~ kw.update(filter=dict(type='auto'))
            #~ kw.update(type=self.filter_type)
        #~ return kw
            
    def get_column_options(self,**kw):
        #~ raise "get_column_options() %s" % self.__class__
        #~ kw.update(xtype='gridcolumn')
        kw.update(dataIndex=self.field.name)
        #~ if self.label is None:
            #~ kw.update(header=self.field.name)
        #~ else:
        kw.update(header=unicode(self.label))
        if not self.editable:
            kw.update(editable=False)
        if not self.sortable:
            kw.update(sortable=False)
        w = self.width or self.preferred_width
        kw.update(width=w*EXT_CHAR_WIDTH)
        return kw    
        
        
    def submit_fields(self):
        return [self.field.name]
        
    def get_field_options(self,**kw):
        if self.xtype:
            kw.update(xtype=self.xtype)
        kw.update(name=self.field.name)
        if self.label:
            kw.update(fieldLabel=unicode(self.label))
        if not self.field.blank:
            kw.update(allowBlank=False)
        if not self.editable:
            kw.update(disabled=True)
            kw.update(readOnly=True)
        #~ http://www.rowlands-bcs.com/extjs/tips/tooltips-form-fields
        #~ if self.field.__doc__:
            #~ kw.update(toolTipText=self.field.__doc__)
        return kw
        
    def ext_options(self,**kw):
        kw = LayoutElement.ext_options(self,**kw)
        kw.update(self.get_field_options())
        return kw
    
        
class TextFieldElement(FieldElement):
    #~ xtype = 'textarea'
    filter_type = 'string'
    vflex = True
    value_template = "new qx.ui.form.TextField().set(%s)"
    xtype = None
    #width = 60
    preferred_width = 60
    #~ preferred_height = 1
    #collapsible = True
    
    #~ def __init__(self,*args,**kw):
        #~ kw.update(defaultAutoCreate = dict(
            #~ tag="textarea",
            #~ autocomplete="off"
        #~ ))
        #~ FieldElement.__init__(self,*args,**kw)

class HtmlTextFieldElement(TextFieldElement):
    pass
    #~ value_template = "new Ext.form.HtmlEditor(%s)"
    #~ # xtype = 'htmleditor'
        
class CharFieldElement(FieldElement):
    filter_type = 'string'
    value_template = "new qx.ui.form.TextField().set(%s)"
    sortable = True
    xtype = None
  
    def __init__(self,*args,**kw):
        FieldElement.__init__(self,*args,**kw)
        self.preferred_width = min(20,max(3,self.field.max_length))
            
    def get_field_options(self,**kw):
        kw = FieldElement.get_field_options(self,**kw)
        kw.update(maxLength=self.field.max_length)
        return kw
        
class FileFieldElement(CharFieldElement):
    #~ value_template = "Lino.file_field_handler(ww,%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    
    #~ def __init__(self,lh,*args,**kw):
        #~ CharFieldElement.__init__(self,lh,*args,**kw)
        #~ lh.has_upload = True
        
    #~ def get_field_options(self,**kw):
        #~ kw = CharFieldElement.get_field_options(self,**kw)
        #~ kw.update(emptyText=_('Select a document to upload...'))
        #~ # kw.update(buttonCfg=dict(iconCls='upload-icon'))
        #~ return kw
    
class ComboFieldElement(FieldElement):
    #~ value_template = "new Ext.form.ComboBox(%s)"        
    sortable = True
    xtype = None
    filter_type = 'string'
    
class ChoicesFieldElement(ComboFieldElement):
    #~ value_template = "new Lino.ChoicesFieldElement(%s)"
    #~ value_template = "new qx.ui.form.ComboBox().set(%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
  
    def __init__(self,*args,**kw):
        FieldElement.__init__(self,*args,**kw)
        self.preferred_width = 20
        
    def get_field_options(self,**kw):
        kw = FieldElement.get_field_options(self,**kw)
        kw.update(store=self.field.choices)
        kw.update(hiddenName=self.field.name+ext_requests.CHOICES_HIDDEN_SUFFIX)
        return kw
        

class RemoteComboFieldElement(ComboFieldElement):
    #~ value_template = "new Lino.RemoteComboFieldElement(%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
  
    def store_options(self,**kw):
        proxy = dict(url=self.lh.rh.ui.get_choices_url(self),method='GET')
        kw.update(proxy=js_code("new Ext.data.HttpProxy(%s)" % py2js(proxy)))
        # a JsonStore without explicit proxy sometimes used method POST
        return kw
      
    def get_field_options(self,**kw):
        kw = FieldElement.get_field_options(self,**kw)
        sto = self.store_options()
        #print repr(sto)
        kw.update(store=js_code("new Lino.ComplexRemoteComboStore(%s)" % py2js(sto)))
        return kw
        
class SimpleRemoteComboFieldElement(RemoteComboFieldElement):
    #~ value_template = "new Lino.SimpleRemoteComboFieldElement(%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    
  
class ComplexRemoteComboFieldElement(RemoteComboFieldElement):
    #~ value_template = "new Lino.ComplexRemoteComboFieldElement(%s)"
        
    def get_field_options(self,**kw):
        kw = RemoteComboFieldElement.get_field_options(self,**kw)
        kw.update(hiddenName=self.field.name+ext_requests.CHOICES_HIDDEN_SUFFIX)
        return kw
        
        
class ForeignKeyElement(ComplexRemoteComboFieldElement):
    
    def __init__(self,lh,field,**kw):
    #~ def __init__(self,*args,**kw):
        #~ print 20100903,repr(self.field.rel.to)
        #~ assert issubclass(self.field.rel.to,models.Model), "%r is not a model" % self.field.rel.to
        self.report = reports.get_model_report(field.rel.to)
        a = self.report.detail_action
        if a is not None:
            #~ self.value_template = "new Lino.TwinCombo(%s)"
            self.value_template = "new qx.ui.basic.Atom().set(%s)"
            kw.update(onTrigger2Click=js_code(
              "function(e){ Lino.show_fk_detail(this,e,Lino.%s)}" % a))
        FieldElement.__init__(self,lh,field,**kw)
      
    def submit_fields(self):
        return [self.field.name,self.field.name+ext_requests.CHOICES_HIDDEN_SUFFIX]
        
        
    def get_field_options(self,**kw):
        kw = ComplexRemoteComboFieldElement.get_field_options(self,**kw)
        kw.update(pageSize=self.report.page_length)
        kw.update(emptyText=_('Select a %s...') % self.report.model._meta.verbose_name)
        return kw



class TimeFieldElement(FieldElement):
    xtype = 'timefield'
    data_type = 'time' # for store column
    sortable = True
    preferred_width = 8
    filter_type = 'time'
  
class DateFieldElement(FieldElement):
    xtype = 'datefield'
    data_type = 'date' # for store column
    sortable = True
    preferred_width = 8
    filter_type = 'date'
    # todo: DateFieldElement.preferred_width should be computed from Report.date_format
    #~ grid_column_template = "new Ext.grid.DateColumn(%s)"
    
    def get_field_options(self,**kw):
        kw = FieldElement.get_field_options(self,**kw)
        kw.update(format=self.lh.rh.report.date_format)
        return kw
        
    def get_column_options(self,**kw):
        kw = FieldElement.get_column_options(self,**kw)
        kw.update(xtype='datecolumn')
        kw.update(format=self.lh.rh.report.date_format)
        return kw
    
class MonthFieldElement(DateFieldElement):
    def get_field_options(self,**kw):
        kw = DateFieldElement.get_field_options(self,**kw)
        kw.update(format='m/Y')
        kw.update(plugins='monthPickerPlugin')
        return kw
        
    
class URLFieldElement(CharFieldElement):
    sortable = True
    preferred_width = 40
    #~ value_template = "new Lino.URLField(%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    
    #~ def get_field_options(self,**kw):
        #~ kw = FieldElement.get_field_options(self,**kw)
        #~ kw.update(vtype='url') #,vtypeText=
        #~ return kw
        
    
class IntegerFieldElement(FieldElement):
    value_template = "new qx.ui.form.TextField().set(%s)"
    filter_type = 'numeric'
    #~ xtype = 'numberfield'
    sortable = True
    preferred_width = 5
    data_type = 'int' 
    def __init__(self,*args,**kw):
        kw.update(textAlign='right')
        FieldElement.__init__(self,*args,**kw)

class DecimalFieldElement(FieldElement):
    filter_type = 'numeric'
    xtype = 'numberfield'
    sortable = True
    data_type = 'float' 
    #~ grid_column_template = "new Ext.grid.NumberColumn(%s)"
    
    def __init__(self,*args,**kw):
        FieldElement.__init__(self,*args,**kw)
        self.preferred_width = min(5,self.field.max_digits) \
                + self.field.decimal_places
                
    def get_column_options(self,**kw):
        kw = FieldElement.get_column_options(self,**kw)
        kw.update(xtype='numbercolumn')
        kw.update(align='right')
        fmt = "0,000"
        if self.field.decimal_places > 0:
            fmt += "." + ("0" * self.field.decimal_places)
        kw.update(format=fmt)
        return kw
        
                

class BooleanFieldElement(FieldElement):
  
    value_template = "new qx.ui.form.CheckBox().set(%s)"
    #~ xtype = 'checkbox'
    data_type = 'boolean' 
    filter_type = 'boolean'
    #~ grid_column_template = "new Ext.grid.BooleanColumn(%s)"
    #~ def __init__(self,*args,**kw):
        #~ FieldElement.__init__(self,*args,**kw)
        
    def set_parent(self,parent):
        FieldElement.set_parent(self,parent)
        #~ if isinstance(parent,Panel) and parent.hideCheckBoxLabels:
        if parent.hideCheckBoxLabels:
            self.update(hideLabel=True)
            
    def get_field_options(self,**kw):
        kw = FieldElement.get_field_options(self,**kw)
        del kw['fieldLabel']
        #~ kw.update(hideLabel=True)
        kw.update(boxLabel=self.label)
        return kw
        
    def get_column_options(self,**kw):
        kw = FieldElement.get_column_options(self,**kw)
        kw.update(xtype='booleancolumn')
        kw.update(trueText=self.lh.rh.report.boolean_texts[0])
        kw.update(falseText=self.lh.rh.report.boolean_texts[1])
        kw.update(undefinedText=self.lh.rh.report.boolean_texts[2])
        return kw
        
    def get_from_form(self,instance,values):
        """
        standard HTML submits checkboxes of a form only when they are checked.
        So if the field is not contained in values, we take False as value.
        """
        instance[self.field.name] = values.get(self.field.name,False)


class DisplayElement(FieldElement):
#~ class ShowOrCreateElement(FieldElement):
    ext_suffix = "_disp"
    declare_type = jsgen.DECLARE_VAR
    #~ value_template = "new Ext.form.DisplayField(%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    #~ value_template = "new Lino.ButtonField(ww,%s)"
    #~ preferred_height = 5
    #~ vflex = True
    #~ filter_type = 'string'
    #~ refers_to_ww = True
    
#~ class QuickActionElement(DisplayElement):
    #~ pass
    #~ ext_suffix = "_quick"
    #~ value_template = "new Lino.QuickAction(ww,%s)"
    
  
  
class HtmlBoxElement(DisplayElement):
    ext_suffix = "_htmlbox"
    #~ declare_type = jsgen.DECLARE_VAR
    #~ value_template = "new Lino.HtmlBoxPanel(ww,%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    preferred_height = 5
    vflex = True
    filter_type = 'string'
    refers_to_ww = True
    
    #~ def __init__(self,lh,name,action,**kw):
        #~ kw.update(plugins=js_code('Lino.HtmlBoxPlugin'))
        #~ LayoutElement.__init__(self,lh,name,**kw)
        
    def get_field_options(self,**kw):
        kw.update(name=self.field.name)
        kw.update(layout='fit')
        if self.field.drop_zone: # testing with drop_zone 'FooBar'
            kw.update(listeners=dict(render=js_code('initialize%sDropZone' % self.field.drop_zone)))
        kw.update(items=js_code("new Ext.BoxComponent()"))
        if self.label:
            kw.update(title=unicode(self.label))
        #~ if self.field.bbar is not None:
            #~ kw.update(ls_bbar_actions=self.field.bbar)
        return kw
        


class Container(LayoutElement):
    vertical = False
    hpad = 1
    is_fieldset = False
    #~ xtype = 'container'
    value_template = "new qx.ui.container.Composite().set(%s)"
    #~ hideCheckBoxLabels = True
    
    #declare_type = jsgen.DECLARE_INLINE
    #declare_type = jsgen.DECLARE_THIS
    declare_type = jsgen.DECLARE_VAR
    #~ declare_type = jsgen.DECLARE_THIS
    
    
    def __init__(self,lh,name,*elements,**kw):
        self.has_frame = lh.layout.has_frame
        self.labelAlign = lh.layout.label_align
        self.hideCheckBoxLabels = lh.layout.hideCheckBoxLabels
        self.active_children = []
        self.elements = elements
        if elements:
            #~ self.has_fields = False
            for e in elements:
                e.set_parent(self)
                #~ if isinstance(e,FieldElement):
                    #~ self.has_fields = True
                if not isinstance(e,LayoutElement):
                    raise Exception("%r is not a LayoutElement" % e)
                if e.active_child:
                    self.active_children.append(e)
                elif isinstance(e,Panel):
                    self.active_children += e.active_children
                    #~ self.has_fields = True
            kw.update(items=elements)
                
        LayoutElement.__init__(self,lh,name,**kw)
        
        #~ if lh.layout.__class__.__name__ == 'CompanyDetail':
            #~ print 20100919, self.__class__.__name__, name, ':', str(elements)
        
        #~ onlyCheckBoxes = True
        #~ for e in self.walk():
            #~ if not isinstance(e,BooleanFieldElement):
                #~ onlyCheckBoxes = False
        #~ if onlyCheckBoxes:
            #~ for e in self.walk():
                #~ e.update(hideLabel=True)
                #~ print 'hideLabel:', e,
        
    def subvars(self):
        return self.elements
            
    def walk(self):
        for e in self.elements:
            for el in e.walk():
                yield el
        yield self
        

    def pprint(self,level=0):
        margin = "  " * level
        s = margin + str(self) + ":\n"
        # self.__class__.__name__
        for e in self.elements:
            for ln in e.pprint(level+1).splitlines():
                s += ln + "\n"
        return s




class Panel(Container):
    """
    A vertical Panel is vflex if and only if at least one of its children is vflex.
    A horizontal Panel is vflex if and only if *all* its children are vflex 
    (if vflex and non-vflex elements are together in a hbox, then the 
    vflex elements will get the height of the highest non-vflex element).
    """
    ext_suffix = "_panel"
    active_child = False
    #~ declare_type = jsgen.DECLARE_VAR
    #~ value_template = "new Ext.Panel(%s)"
    
    def __init__(self,lh,name,vertical,*elements,**kw):
        self.vertical = vertical
            
        #~ if self.vertical:
            #~ vflex_elems = [e for e in elements if e.vflex]
            #~ if len(flex_elems) > 0:
                #~ assert len(kw) == 0, "%r is not empty" % kw
                #~ vfix_elems = [e for e in elements if not e.vflex]
                #~ flex_panel = Panel(lh,name,vertical,*vflex_elems)
                #~ fix_panel = Panel(lh,name,vertical,*vfix_elems)
                
        self.vflex = not vertical
        stretch = False
        #~ monitorResize = False
        for e in elements:
            #~ if e.collapsible:
                #~ monitorResize = True
            if e.vflex:
                stretch = True
            if self.vertical:
                if e.vflex:
                    self.vflex = True
            else:
                if not e.vflex:
                    self.vflex = False
                    #~ print 20100615, self.lh.layout, self, "hbox loses vflex because of", e
                    
        if len(elements) > 1 and self.vflex:
            if self.vertical:
                """
                Example : The panel contains a mixture of fields and grids. 
                Fields are not vflex, grids well.
                """
                #~ print 20100615, self.lh, self
                # so we must split this panel into several containers.
                # vflex elements go into a vbox, the others into a form layout. 
                
                if False:
                    # Rearrange elements into "element groups"
                    # Each element of egroups is a list of elements who have same vflex
                    egroups = []
                    for e in elements:
                        if len(egroups) and egroups[-1][0].vflex == e.vflex:
                            egroups[-1].append(e)
                        else:
                            egroups.append([e])
                            
                    if len(egroups) == 1:
                        # all elements have same vflex
                        assert tuple(egroups[0]) == elements, "%r != %r" % (tuple(egroups[0]), elements)
                        
                    elements = []
                    for eg in egroups:
                        if eg[0].vflex:
                            #~ for e in eg: e.update(flex=1,align='stretch')
                            for e in eg: 
                                e.update(flex=1)
                                e.collapsible = True
                                #~ e.update(collapsible=True)
                            if len(eg) == 1:
                                g = eg[0]
                            else:
                                #~ g = Container(lh,name,*eg,**dict(layout='vbox',flex=1)
                                g = Panel(lh,name,vertical,*eg,**dict(layout='vbox',
                                  flex=1,layoutConfig=dict(align='stretch')))
                                assert g.vflex is True
                        else:
                            #~ for e in eg: e.update(align='stretch')
                            if len(eg) == 1:
                                g = eg[0]
                            else:
                                g = Container(lh,name,*eg,**dict(layout='form',autoHeight=True))
                                #~ g = Container(lh,name,*eg,**dict(layout='form'))
                                assert g.vflex is False
                        #~ if monitorResize:
                            #~ g.update(monitorResize=True)
                        #~ g.update(align='stretch')
                        #~ g.update(layoutConfig=dict(align='stretch'))
                        elements.append(g)
                    kw.update(layout='vbox',layoutConfig=dict(align='stretch'))
                    #~ self.elements = elements
            else: # not self.vertical
                kw.update(layout='hbox',layoutConfig=dict(align='stretch'))
              
        for e in elements:
            #~ e.set_parent(self)
            if isinstance(e,FieldElement):
                self.is_fieldset = True
                #~ if self.label_width < e.label_width:
                    #~ self.label_width = e.label_width
                if e.label:
                    w = len(e.label) + 1 # +1 for the ":"
                    if self.label_width < w:
                        self.label_width = w


        label = lh.layout.collapsible_elements.get(name,None)
        if label:
            self.collapsible = True
            self.label = label

        Container.__init__(self,lh,name,*elements,**kw)
        

        w = h = 0
        for e in self.elements:
            ew = e.width or e.preferred_width
            eh = e.height or e.preferred_height
            if self.vertical:
                #~ h += e.flex
                h += eh
                w = max(w,ew)
            else:
                #w += e.flex
                w += ew
                h = max(h,eh)
        self.preferred_height = h
        self.preferred_width = w
        assert self.preferred_height > 0, "%s : preferred_height is 0" % self
        assert self.preferred_width > 0, "%s : preferred_width is 0" % self
        
        d = self.value
        if not d.has_key('layout'):
            if len(self.elements) == 1:
                d.update(layout='fit')
            elif self.vertical:
                #~ d.update(layout='form')
                if self.vflex:
                    d.update(layout='vbox',layoutConfig=dict(align='stretch'))
                else:
                    # 20100921b
                    #~ d.update(layout='form')
                    d.update(layout='form',autoHeight=True)
                    #~ d.update(layout='vbox',autoHeight=True)
            else:
                d.update(layout='hbox',autoHeight=True) # 20101028
                #~ d.update(layout='hbox')
                #~ 20101116 d.update(layoutConfig=dict(align='stretchmax'))
                #~ if stretch : # 20100912
                    #~ d.update(layoutConfig=dict(align='stretch'))
                #~ else:
                    #~ d.update(layoutConfig=dict(align='stretchmax'))
                
        if d['layout'] == 'form':
            assert self.vertical
            #~ d.update(autoHeight=True)
            if len(self.elements) == 1 and self.elements[0].vflex:
                self.elements[0].update(anchor="100% 100%")
            else:
                for e in self.elements:
                    #~ assert not e.vflex
                    #~ h = e.height or e.preferred_height
                    #~ if e.vflex:
                        #~ e.value.update(anchor="100% 100%")
                    #~ else:
                        #~ e.value.update(anchor="100%")
                    e.update(anchor="100%")
                
        if d['layout'] == 'hbox':
            if not self.vflex: # 20101028
                d.update(autoHeight=True)
                d.update(layoutConfig=dict(align='stretchmax'))
            for e in self.elements:
                if e.hflex:
                    w = e.width or e.preferred_width
                    #~ e.value.update(columnWidth=float(w)/self.preferred_width) # 20100615
                    e.value.update(flex=int(w*100/self.preferred_width))
                
            #~ wrappers = []
            #~ for e in self.elements:
                #~ wrappers.append(dict(layout='form',autoHeight=True,items=e))
            #~ d.update(items=wrappers)
              
        elif d['layout'] == 'vbox':
            "a vbox with 2 or 3 elements, of which at least two are vflex will be implemented as a VBorderPanel"
            assert len(self.elements) > 1
            vflex_count = 0
            h = self.height or self.preferred_height
            for e in self.elements:
                eh = e.height or e.preferred_height
                if e.vflex:
                    e.value.update(flex=int(eh*100/h))
                    vflex_count += 1
            if vflex_count >= 2 and len(self.elements) <= 3:
            #~ if vflex_count >= 1 and len(self.elements) <= 3:
                self.remove('layout','layoutConfig')
                #~ self.value_template = 'new Lino.VBorderPanel(%s)'
                for e in self.elements:
                    #~ if e.vflex: # """as long as there are bugs, make also non-vflex resizable"""
                    e.update(split=True)
                self.elements[0].update(region='north')
                self.elements[1].update(region='center')
                if len(self.elements) == 3:
                    self.elements[1].update(region='south')
        
        

class GridElement(Container): 
    declare_type = jsgen.DECLARE_VAR
    #~ value_template = "new lino.TablePanel(ww,%s)"
    value_template = "new qx.ui.basic.Atom().set(%s)"
    ext_suffix = "_grid"
    vflex = True
    xtype = None
    preferred_height = 5
    refers_to_ww = True
    
    def __init__(self,lh,name,rpt,*elements,**kw):
        """
        :param lh: the handle of the DetailLayout owning this grid
        :param rpt: the report being displayed
        """
        assert isinstance(rpt,reports.Report), "%r is not a Report!" % rpt
        self.report = rpt
        if len(elements) == 0:
            self.rh = rpt.get_handle(lh.rh.ui)
            elements = self.rh.list_layout._main.elements
            #~ columns = self.rh.list_layout._main.elements
        w = 0
        for e in elements:
            w += (e.width or e.preferred_width)
        self.preferred_width = constrain(w,10,120)
        #~ kw.update(boxMinWidth=500)
        #~ self.columns = elements
        #~ self.store = qx_store.Store(lh)
        self.columns = [GridColumn(i,e) for i,e in enumerate(elements)]
        
        kw.update(page_length=self.report.page_length)

        a = rpt.get_action('detail')
        if a:
            kw.update(ls_detail_handler=js_code("Lino.%s" % a))
        a = rpt.get_action('insert')
        if a:
            kw.update(ls_insert_handler=js_code("Lino.%s" % a))
        
        Container.__init__(self,lh,name,*elements,**kw)
        #~ self.active_children = columns
        assert not kw.has_key('before_row_edit')
        self.update(before_row_edit=before_row_edit(self))
        
        self.preferred_height = rpt.page_length 
        if self.report.master is not None:
            self.mt = ContentType.objects.get_for_model(self.report.master).pk
        else:
            self.mt = 'undefined'
            
            
    def update_config(self,wc):
        pass
        #~ for i,w in enumerate(wc['column_widths']):
            #~ self.column_model.columns[i].update(width = w)

    def ext_options(self,**kw):
        rh = self.report.get_handle(self.lh.rh.ui)
        kw = LayoutElement.ext_options(self,**kw)
        #~ d.update(ls_data_url=rh.ui.get_actor_url(self.report))
        kw.update(ls_url=rh.ui.build_url(self.report.app_label,self.report._actor_name))
        kw.update(ls_store_fields=[js_code(f.as_js()) for f in rh.store.list_fields])
        #~ kw.update(ls_columns=[GridColumn(i,e) for i,e in enumerate(self.columns)])
        #~ kw.update(ls_filters=[e.get_filter_options() for e in self.elements if e.filter_type])
        kw.update(ls_id_property=rh.store.pk.name)
        kw.update(pk_index=rh.store.pk_index)
        kw.update(ls_quick_edit=rh.report.cell_edit)
        kw.update(ls_bbar_actions=[rh.ui.a2btn(a) for a in rh.get_actions(rh.report.default_action)])
        kw.update(ls_grid_configs=[gc.data for gc in self.report.grid_configs])
        kw.update(gc_name=DEFAULT_GC_NAME)
        #~ gc = self.report.grid_configs.get('',None)
        #~ if gc is not None:
            #~ kw.update(ls_grid_config=gc)
        return kw
        
class SlaveGridElement(GridElement):
    def ext_options(self,**kw):
        kw = GridElement.ext_options(self,**kw)
        #~ kw.update(plugins=js_code('new Lino.SlaveGridPlugin(caller)'))
        
        kw.update(title=unicode(self.report.label))
        
        #~ js = "Lino.do_action(caller,%r)" % \
            #~ rh.list_layout.get_absolute_url(run=True)
        #~ js = "function(){ Lino.notify('Das ist noch nicht fertig... siehe ext_elems.py:990');}"
        #~ kw.update(listeners=dict(click=js_code(js)))
        return kw
      
        
class M2mGridElement(GridElement):
    def __init__(self,lh,field,*columns,**kw):
        self.field = field
        rpt = reports.get_model_report(field.rel.to)
        #~ rh = rpt.get_handle(lh.ui)
        GridElement.__init__(self,lh,id2js(rpt.actor_id),rpt,*columns,**kw)
  

            
class MainPanel(jsgen.Variable):
    declare_type = jsgen.DECLARE_INLINE
    refers_to_ww = True
  
    def apply_window_config(self,wc):
        pass
  
    def setup(self):
        pass
                
    @classmethod
    def field2elem(cls,lh,field,**kw):
        #~ if hasattr(field,'_lino_chooser'):
        ch = choosers.get_for_field(field)
        if ch:
            if ch.simple_values:
                return SimpleRemoteComboFieldElement(lh,field,**kw)
            else:
                if isinstance(field,models.ForeignKey):
                    return ForeignKeyElement(lh,field,**kw)
                else:
                    return ComplexRemoteComboFieldElement(lh,field,**kw)
        if field.choices:
            return ChoicesFieldElement(lh,field,**kw)
            
        if isinstance(field,fields.VirtualField):
            field = field.return_type
        for cl,x in _field2elem:
            if isinstance(field,cl):
                return x(lh,field,**kw)
        raise NotImplementedError("No LayoutElement for %s" % field.__class__)
        #~ raise NotImplementedError("field %r" % field)



class GridMainPanel(GridElement,MainPanel):
    def __init__(self,lh,name,vertical,*columns,**kw):
        """ignore the "vertical" arg"""
        self.value_template = "new lino.%s(%%s)" % lh.rh.report
        GridElement.__init__(self,lh,name,lh.rh.report,*columns,**kw)
        



class DetailMainPanel(Panel,MainPanel):
#~ class DetailMainPanel(Panel,WrappingMainPanel):
    #~ declare_type = jsgen.DECLARE_THIS
    #~ xtype = 'form'
    xtype = None
    #~ value_template = "new Ext.form.FormPanel(%s)"
    value_template = "new Ext.Panel(%s)"
    def __init__(self,lh,name,vertical,*elements,**kw):
        #~ self.rh = lh.datalink
        self.report = lh.rh.report
        #~ MainPanel.__init__(self)
        #~ DataElementMixin.__init__(self,lh.link)
        kw.update(autoScroll=True)
        #~ kw.update(height=800, autoScroll=True)
        Panel.__init__(self,lh,name,vertical,*elements,**kw)
        #lh.needs_store(self.rh)
        
        
    def unused_get_datalink(self):
        return self.rh
        
    def subvars(self):
        #~ print 'DetailMainPanel.subvars()', self
        for e in MainPanel.subvars(self):
            yield e
        for e in Panel.subvars(self):
            yield e
            
    def ext_options(self,**kw):
        self.setup()
        kw = Panel.ext_options(self,**kw)
        if self.lh.layout.label:
            #~ kw.update(title=unicode(self.lh.layout.label))
            kw.update(title=_(self.lh.layout.label))
        #d.update(region='east',split=True) #,width=300)
        #~ kw.update(width=800)
        #~ kw.update(autoScroll=True)
        if False:
            kw.update(tbar=js_code("""new Ext.PagingToolbar({
              store: %s,
              displayInfo: true,
              pageSize: 1,
              prependButtons: true,
            }) """ % self.rh.store.ext_name))
        #d.update(items=js_code(self._main.as_ext(request)))
        #d.update(items=js_code("[%s]" % ",".join([e.as_ext() for e in self.elements])))
        #~ kw.update(items=self.elements)
        #d.update(autoHeight=False)
        #~ kw.update(bbar=self.bbar_buttons)
        #~ kw.update(bbar=self.buttons)
        #d.update(standardSubmit=True)
        return kw
        
    #~ @classmethod
    #~ def field2elem(cls,lh,field,**kw):
        #~ e = MainPanel.field2elem(lh,field,**kw)
        #~ if isinstance(e,HtmlBoxElement): return e
        #~ po = dict(layout='form',autoHeight=True) # 20101028
        #~ ct = Panel(lh,field.name+"_ct",True,e,**po)#,flex=0)
        #~ ct.field = field
        #~ return ct

class TabPanel(jsgen.Component):
#~ class TabPanel(jsgen.Value):
    value_template = "new Ext.TabPanel(%s)"

        
    def __init__(self,tabs,**kw):
        self.active_children = []
        for t in tabs:
            self.active_children += t.active_children
            t.update(listeners=dict(activate=js_code("Lino.on_tab_activate")))
            #~ if t.has_upload:
                #~ self.has_upload = True
      
        self.tabs = tabs
        kw.update(
          split=True,
          activeTab=0,
          #~ layoutOnTabChange=True, # 20101028
          #~ forceLayout=True, # 20101028
          #~ deferredRender=False, # 20101028
          #~ autoScroll=True, 
          #~ width=300, # ! http://code.google.com/p/lino/wiki/20100513
          items=tabs,
          # http://www.extjs.com/forum/showthread.php?26564-Solved-FormPanel-in-a-TabPanel
          #~ listeners=dict(activate=js_code("function(p) {p.doLayout();}"),single=True),
        )
        jsgen.Value.__init__(self,kw)
        
    def has_field(self,f):
        for t in self.tabs:
            if t.has_field(f): 
                return True



class FormPanel(jsgen.Component):
    declare_type = jsgen.DECLARE_VAR
    #~ value_template = "new Lino.FormPanel(ww,%s)"
    listeners = None
    
    def __init__(self,rh,action,main,**kw):
        self.rh = rh
        self.value_template = "new Lino.%s.FormPanel(ww,%%s)" % self.rh.report
        self.main = main
        kw.update(
          #~ items=main,
          #~ autoScroll=True,
          #~ autoHeight=True,
          layout='fit',
        )
        if not isinstance(action,reports.InsertRow):
            kw.update(has_navigator=rh.report.has_navigator)
            
        on_render = []
        elems_by_field = {}
        field_elems = []
        for e in main.active_children:
            if isinstance(e,FieldElement):
                field_elems.append(e)
                l = elems_by_field.get(e.field.name,None)
                if l is None:
                    l = []
                    elems_by_field[e.field.name] = l
                l.append(e)
            
        for e in field_elems:
            #~ if isinstance(e,FileFieldElement):
                #~ kw.update(fileUpload=True)
            chooser = choosers.get_for_field(e.field)
            if chooser:
                #~ logger.debug("20100615 %s.%s has chooser", self.lh.layout, e.field.name)
                for f in chooser.context_fields:
                    for el in elems_by_field.get(f.name,[]):
                        #~ if main.has_field(f):
                        #~ varname = varname_field(f)
                        #~ on_render.append("%s.on('change',Lino.chooser_handler(%s,%r));" % (varname,e.ext_name,f.name))
                        on_render.append(
                            "%s.on('change',Lino.chooser_handler(%s,%r));" % (
                            el.ext_name,e.ext_name,f.name))
        
        if on_render:
            assert not kw.has_key('listeners')
            #~ kw.update(listeners=dict(render=js_code('function(){%s}' % '\n'.join(on_render))))
            self.listeners=dict(render=js_code('function(){%s}' % '\n'.join(on_render)))
        #~ kw.update(before_row_edit=before_row_edit(main))
        self.before_row_edit=before_row_edit(main)
        
        rpt = rh.report
        a = rpt.get_action('detail')
        if a:
            kw.update(ls_detail_handler=js_code("Lino.%s" % a))
        a = rpt.get_action('insert')
        if a:
            kw.update(ls_insert_handler=js_code("Lino.%s" % a))
        
        kw.update(ls_bbar_actions=[rh.ui.a2btn(a) for a in rpt.get_actions(action)])
        kw.update(ls_url=rh.ui.build_url(rpt.app_label,rpt._actor_name))
        jsgen.Component.__init__(self,'form_panel',**kw)
        
    def has_field(self,f):
        return self.main.has_field(f)



_field2elem = (
    (fields.HtmlBox, HtmlBoxElement),
    #~ (fields.QuickAction, QuickActionElement),
    (fields.DisplayField, DisplayElement),
    (models.URLField, URLFieldElement),
    (models.FileField, FileFieldElement),
    (models.EmailField, CharFieldElement),
    (fields.HtmlTextField, HtmlTextFieldElement),
    (models.TextField, TextFieldElement),
    (models.CharField, CharFieldElement),
    (fields.MonthField, MonthFieldElement),
    (models.DateField, DateFieldElement),
    #~ (models.TimeField, TimeFieldElement),
    (models.IntegerField, IntegerFieldElement),
    (models.DecimalField, DecimalFieldElement),
    (models.BooleanField, BooleanFieldElement),
    (models.ManyToManyField, M2mGridElement),
    (models.ForeignKey, ForeignKeyElement),
    (models.AutoField, IntegerFieldElement),
)
    




# -*- coding: UTF-8 -*-
## Copyright 2009-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
Performs a database flush, removing *all existing tables* from the 
database (not only Django tables), then runs Django's standard `syncdb` 
and `loaddata` commands to load the specified fixtures for all applications.
Also writes log entries to your dblogger.

That may sound dangerous, but that's what we want when we have a 
:doc:`python dump </topics/dumpy>` to restore our database. You know that you 
should rather not let Lino and some other application share the same database!

Django's `reset` command may fail after an upgrade if the new Lino 
version defines new tables. In that case, flush sends a DROP TABLE 
which fails because that table doesn't exist. 

Lino's `initdb` reimplements a simplified version of Django's `reset` command, 
but without the possibility of deleting only *some* data (the thing which 
caused so big problems that Django 1.3. decided to `deprecate this command
<https://docs.djangoproject.com/en/dev/releases/1.3/#reset-and-sqlreset-management-commands>`__.

See also ticket :doc:`/tickets/50`.

"""

import logging
from optparse import make_option 

from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError

from django.core.management.sql import sql_delete, sql_flush
from django.core.management.color import no_style
#~ from django.core.management.sql import sql_reset
from django.db import connections, transaction, DEFAULT_DB_ALIAS
from django.db import models

import lino
from lino.core.modeltools import app_labels
from lino.utils import *

USE_SQLDELETE = True

USE_DROP_CREATE = False # tried, but doesn't seem to work
"""
http://stackoverflow.com/questions/3414247/django-drop-all-tables-from-database
http://thingsilearned.com/2009/05/10/drop-database-command-for-django-manager/
"""

class Command(BaseCommand):
    help = __doc__
    
    args = "fixture [fixture ...]"
    
    option_list = BaseCommand.option_list + (
        make_option('--noinput', action='store_false', dest='interactive', default=True,
            help='Do not prompt for input of any kind.'),
        make_option('--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS, help='Nominates a database to reset. '
                'Defaults to the "default" database.'),
    ) 

    def handle(self, *args, **options):
      
        from lino.core.kernel import analyze_models
        analyze_models()
      
        from lino.utils import dblogger
            
        if not dblogger.logger.isEnabledFor(logging.INFO):
            raise CommandError("System logger must be enabled for INFO")
        using = options.get('database', DEFAULT_DB_ALIAS)
        dbname = settings.DATABASES[using]['NAME']
        if options.get('interactive'):
            if not confirm("We are going to flush your database (%s).\nAre you sure (y/n) ?" % dbname):
                raise CommandError("User abort.")
            
        options.update(interactive=False)
        dblogger.info("Lino initdb %s started on database %s.", args, dbname)
        dblogger.info(settings.LINO.welcome_text())
        
        if USE_DROP_CREATE:
        
            from django.db import connection
            cursor = connection.cursor()
            #~ cursor.execute("DROP DATABASE %s;", [connection.settings_dict['NAME']])
            #~ cursor.execute("CREATE DATABASE %s;", [connection.settings_dict['NAME']])        
            cursor.execute("DROP DATABASE %s;" % dbname)
            cursor.execute("CREATE DATABASE %s;" % dbname)
                
        else:
        
            sql_list = []
            conn = connections[using]
            
            if USE_SQLDELETE:
                #~ sql_list = u'\n'.join(sql_reset(app, no_style(), conn)).encode('utf-8')
                app_list = [models.get_app(app_label) for app_label in app_labels()]
                for app in app_list:
                    # app_label = app.__name__.split('.')[-2]
                    sql_list.extend(sql_delete(app,no_style(),conn))
                    # print app_label, ':', sql_list
            else:
                #~ call_command('flush',verbosity=0,interactive=False)
                #~ call_command('flush',**options)
                sql_list.extend(sql_flush(no_style(), conn, only_django=False))
                
            #~ print sql_list
            
            try:
                cursor = conn.cursor()
                for sql in sql_list:
                    # print sql
                    cursor.execute(sql)
            except Exception, e:
                transaction.rollback_unless_managed(using=using)
                raise
                
            transaction.commit_unless_managed()
        
        #~ call_command('reset',*apps,**options)
        #~ call_command('syncdb',load_initial_data=False,**options)
        #~ if USE_SQLDELETE:
        
        #~ tried to call `syncdb` with `verbosity=0` to avoid the 
        #~ irritating message "No fixtures found" (which comes because there 
        #~ are no `initial_data` fixtures):
        #~ syncdb_options = dict(**options)
        #~ syncdb_options.update(verbosity=0)
        #~ call_command('syncdb',**syncdb_options)
        #~ not good because all other messages "Creating table..." also disappear.
        
        
        call_command('syncdb',load_initial_data=False,**options)
        
        call_command('loaddata',*args,**options)
        
        dblogger.info("Lino initdb done %s on database %s.", args, dbname)

#~ print 20120426, 'ok'



from odf.style import Style, ParagraphProperties, TextProperties

def addStandardStyles(doc):
    style = Style(name="Standard", family="paragraph", attributes={'class':"text"})
    doc.styles.addElement(style)

    style = Style(name="Text_20_body", displayname="Text body", family="paragraph", parentstylename="Standard", attributes={'class':"text"})
    p = ParagraphProperties(margintop="0cm", marginbottom="0.212cm")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="List_20_Contents", displayname="List Contents", family="paragraph", parentstylename="Standard", attributes={'class':"html"})
    p = ParagraphProperties(marginleft="1cm", marginright="0cm", textindent="0cm", autotextindent="false")
    style.addElement(p)
    doc.styles.addElement(style)

    
    style = Style(name="List_20_Heading", displayname="List Heading", family="paragraph", parentstylename="Standard",
         nextstylename="List_20_Contents", attributes={'class':"html"})
    p = ParagraphProperties(marginleft="0cm", marginright="0cm", textindent="0cm", autotextindent="false")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Text_20_body_20_indent", displayname="Text body indent", family="paragraph", parentstylename="Text_20_body", attributes={'class':"text"})
    p = ParagraphProperties(marginleft="0.499cm", marginright="0cm", textindent="0cm", autotextindent="false")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Heading", family="paragraph", parentstylename="Standard", nextstylename="Text_20_body", attributes={'class':"text"})
    p = ParagraphProperties(margintop="0.423cm", marginbottom="0.212cm", keepwithnext="always")
    style.addElement(p)
    p = TextProperties(fontname="Nimbus Sans L", fontsize="14pt", fontnameasian="DejaVu LGC Sans", fontsizeasian="14pt", fontnamecomplex="DejaVu LGC Sans", fontsizecomplex="14pt")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Heading_20_1", displayname="Heading 1", family="paragraph", parentstylename="Heading", nextstylename="Text_20_body", attributes={'class':"text"}, defaultoutlinelevel=1)
    p = TextProperties(fontsize="115%", fontweight="bold", fontsizeasian="115%", fontweightasian="bold", fontsizecomplex="115%", fontweightcomplex="bold")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Heading_20_2", displayname="Heading 2", family="paragraph", parentstylename="Heading", nextstylename="Text_20_body", attributes={'class':"text"}, defaultoutlinelevel=2)
    p = TextProperties(fontsize="14pt", fontstyle="italic", fontweight="bold", fontsizeasian="14pt", fontstyleasian="italic", fontweightasian="bold", fontsizecomplex="14pt", fontstylecomplex="italic", fontweightcomplex="bold")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Heading_20_3", displayname="Heading 3", family="paragraph", parentstylename="Heading", nextstylename="Text_20_body", attributes={'class':"text"}, defaultoutlinelevel=3)
    p = TextProperties(fontsize="14pt", fontweight="bold", fontsizeasian="14pt", fontweightasian="bold", fontsizecomplex="14pt", fontweightcomplex="bold")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="List", family="paragraph", parentstylename="Text_20_body", attributes={'class':"list"})
    doc.styles.addElement(style)

    style = Style(name="Caption", family="paragraph", parentstylename="Standard", attributes={'class':"extra"})
    p = ParagraphProperties(margintop="0.212cm", marginbottom="0.212cm", numberlines="false", linenumber="0")
    style.addElement(p)
    p = TextProperties(fontsize="12pt", fontstyle="italic", fontsizeasian="12pt", fontstyleasian="italic", fontsizecomplex="12pt", fontstylecomplex="italic")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Index", family="paragraph", parentstylename="Standard", attributes={'class':"index"})
    p = ParagraphProperties(numberlines="false", linenumber=0)
    doc.styles.addElement(style)

    style = Style(name="Source_20_Text", displayname="Source Text", family="text")
    p = TextProperties(fontname="Courier", fontnameasian="Courier", fontnamecomplex="Courier")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Variable", family="text")
    p = TextProperties(fontstyle="italic", fontstyleasian="italic", fontstylecomplex="italic")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Emphasis", family="text")
    p = TextProperties(fontstyle="italic", fontstyleasian="italic", fontstylecomplex="italic")
    style.addElement(p)
    doc.styles.addElement(style)

    style = Style(name="Strong_20_Emphasis", displayname="Strong Emphasis", family="text")
    p = TextProperties(fontweight="bold", fontweightasian="bold", fontweightcomplex="bold")
    style.addElement(p)
    doc.styles.addElement(style)

# Automatic styles
    style = Style(name="Bold", displayname="Bold", family="text")
    p = TextProperties(fontweight="bold", fontweightasian="bold", fontweightcomplex="bold")
    style.addElement(p)
    doc.automaticstyles.addElement(style)

    style = Style(name="Italic", family="text")
    p = TextProperties(fontstyle="italic", fontstyleasian="italic", fontstylecomplex="italic")
    style.addElement(p)
    doc.automaticstyles.addElement(style)




## Copyright 2010-2011 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.


"""
:mod:`lino.utils.dblogger` -- Logging database changes
------------------------------------------------------

This module contains utilities for logging changes in a 
Django database.
Since logging of database changes will inevitably cause some extra work, 
this feature is optional per site and per model.
"""

import logging
logger = logging.getLogger(__name__)
info = logger.info
warning = logger.warning
exception = logger.exception
error = logger.error
debug = logger.debug
#~ getLevel = logger.getLevel
#~ setLevel = logger.setLevel

from lino.core.modeltools import obj2str

def log_deleted(request,elem):
    #~ on_user_change(request,elem)
    logger.info(u"%s deleted by %s.",obj2str(elem),request.user)
        
if False:



    class DiffingMixin(object):
        """
        Unmodified copy of http://djangosnippets.org/snippets/1683/
        
        """
        def __init__(self, *args, **kwargs):
            super(DiffingMixin, self).__init__(*args, **kwargs)
            self._original_state = dict(self.__dict__)
            
        def save(self, *args, **kwargs):
            #~ for name,old_new in self.changed_columns().items():
                
            state = dict(self.__dict__)
            del state['_original_state']
            self._original_state = state
            super(DiffingMixin, self).save()
            
        def is_dirty(self):
            missing = object()
            result = {}
            for key, value in self._original_state.iteritems():
                if value != self.__dict__.get(key, missing):
                    return True
            return False
            
        def changed_columns(self):
            missing = object()
            result = {}
            for key, value in self._original_state.iteritems():
                if value != self.__dict__.get(key, missing):
                    result[key] = {'old':value, 'new':self.__dict__.get(key, missing)}
            return result



    #~ def on_user_change(request,elem):    
        #~ m = getattr(elem,'on_user_change',None)
        #~ if m: 
            #~ m(request)

    def log_created(request,elem):
        #~ on_user_change(request,elem)
        logger.info(u"%s created by %s.",obj2str(elem),request.user)
        
    def log_changes(request,elem):
        """
        Log the changes that have been made to every field of `elem` 
        if `elem` is an instance of `DiffingMixin`, otherwise does nothing.
        """
        #~ on_user_change(request,elem)
        
        if isinstance(elem,DiffingMixin):
            changes = []
            for k,v in elem.changed_columns().items():
                changes.append(u"%s : %s --> %s" % (k,obj2str(v['old']),obj2str(v['new'])))
            if len(changes) == 0:
                changes = '(no changes)'
            #~ elif len(changes) == 1:
                #~ changes = changes[0]
            else:
                changes = '\n- ' + ('\n- '.join(changes))
            msg = u"%s modified by %s : %s" % (
                obj2str(elem),
                request.user,
                changes)
            #~ print msg
            logger.info(msg)

    class PseudoRequest:
        def __init__(self,user):
            self.user = user



## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
A :term:`minimal application` that uses only the 
:mod:`lino.modlib.contacts` module.
"""

#~ from django.db import models
#~ from django.utils.translation import ugettext_lazy as _

#~ from lino import dd

#~ from lino.modlib.contacts import models as contacts

#~ class Person(contacts.Person,contacts.Born):
    #~ class Meta(contacts.Person.Meta):
        #~ app_label = 'contacts'

#~ class Company(contacts.Partner,contacts.CompanyMixin):
    #~ class Meta(contacts.CompanyMixin.Meta):
        #~ app_label = 'contacts'
        

#~ def setup_master_menu(site,ui,user,m):
    #~ m.add_action(site.modules.contacts.Persons)
    #~ m.add_action(site.modules.contacts.Companies)



## Copyright 2008-2010 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

from django.shortcuts import render_to_response
from django.shortcuts import get_object_or_404
#from django.http import HttpResponse
from django.core.paginator import Paginator, InvalidPage, EmptyPage
from lino.modlib.voc import models


def unit_list(request):
    """
    """
    return render_to_response("voc/unit_list.html", 
         dict(units=models.Unit.objects.all()))
      
#~ def course_list(request):
    #~ """
    #~ """
    #~ return render_to_response("voc/course_list.html", 
         #~ dict(courses=models.Course.objects.all()))
      
def unit_detail(request, unit_id):
    """
    """
    unit = get_object_or_404(models.Unit,pk=unit_id)
    #~ try:
        #~ unit = models.Unit.get(pk=unit_id)
    #~ except ObjectDoesNotExist:
        #~ raise Http404, "%s : no such unit" % (pk)
    
    return render_to_response("voc/unit_detail.html", 
          dict(unit=unit))
      
#~ def course_detail(request, course_id):
    #~ """
    #~ """
    #~ course = get_object_or_404(models.Course,pk=course_id)
    #~ return render_to_response("voc/course_detail.html", 
          #~ dict(course=course))
      

      
def entry_page(request, unit_id, page):
    """
    """
    unit = get_object_or_404(models.Unit,pk=unit_id)
    paginator = Paginator(unit.entry_set.all(), 1) # 1 Entry per page
    #~ if not page:
        #~ page = request.GET.get('page', "1")
    #~ try:
        #~ page = int(page)
    #~ except ValueError:
        #~ page=1
    try:
        page_obj = paginator.page(page)
    except (EmptyPage, InvalidPage):
        page_obj = paginator.page(paginator.num_pages)
    #assert len(page_object.object_list) == 1
    entry=page_obj.object_list[0]
    return render_to_response('voc/entry_page.html', 
      dict(entry=entry, page_obj=page_obj))

      

def entry_detail(request, pk):
    """
    """
    obj = get_object_or_404(models.Entry,pk=pk)
    #~ try:
        #~ obj = models.Entry.get(pk=pk)
    #~ except ObjectDoesNotExist:
        #~ raise Http404, "%s : no such entry" % (pk)
    
    return render_to_response("voc/entry_detail.html", 
          dict(entry=obj))



## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
:mod:`lino.core.changes` -- Watching database changes
------------------------------------------------------

This module contains utilities for logging changes in a Django database.

See also :doc:`/topics/changes`


"""

import logging
logger = logging.getLogger(__name__)
info = logger.info
warning = logger.warning
exception = logger.exception
error = logger.error
debug = logger.debug
#~ getLevel = logger.getLevel
#~ setLevel = logger.setLevel

import datetime

from django.conf import settings

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _


from lino.core.modeltools import obj2str, full_model_name
from lino.core import fields 
from lino.core import actions
from lino.core import choicelists

#~ WATCH_SPECS = dict()

NOT_GIVEN = object()

class PseudoRequest:
    def __init__(self,username):
        self.username = username
        self._user = None
        
    def get_user(self):
        if self._user is None:
            if settings.LINO.user_model is not None:
                self._user = settings.LINO.user_model.objects.get(username=self.username)
        return self._user
    user = property(get_user)
        
class ChangeTypes(choicelists.ChoiceList):
    app_label = 'lino'
    verbose_name = _("Change Type")
    verbose_name_plural = _("Change Types")
add = ChangeTypes.add_item    
add('C',_("Create"),'create')
add('U',_("Update"),'update')
add('D',_("Delete"),'delete')
add('R',_("Remove child"),'remove_child')
add('A',_("Add child"),'add_child')

        

class WatcherSpec:
    #~ def __init__(self,ignored_fields,master_key):
    def __init__(self,ignored_fields,get_master):
        self.ignored_fields = ignored_fields
        #~ self.master_key = master_key
        self.get_master = get_master
    
def add_watcher_spec(model,ignore=[],master_key=None,**options):
    #~ if ignore is None:
        #~ model._change_watcher_spec = None
        #~ return
    #~ from lino import dd
    if isinstance(ignore,basestring):
        ignore = fields.fields_list(model,ignore)
    if isinstance(master_key,basestring):
        fld = fields.get_data_elem(model,master_key)
        if fld is None:
            raise Exception("No field %r in %s" % (master_key,model))
        master_key = fld
    if isinstance(master_key,fields.RemoteField):
        get_master = master_key.func
    elif master_key is None:
        def get_master(obj):
            return obj
    else:
        def get_master(obj):
            return getattr(obj,master_key.name)
    ignore = set(ignore)
    #~ cs = WATCH_SPECS.get(model)
    cs = model._change_watcher_spec
    if cs is not None:
        ignore |= cs.ignored_fields
    for f in model._meta.fields:
        if not f.editable:
            ignore.add(f.name)
    model._change_watcher_spec = WatcherSpec(ignore,get_master)
    #~ WATCH_SPECS[model] = WatcherSpec(ignore,master_key)
    
    #~ logger.info("20120924 %s ignore %s", model, ignore)
    #~ model._watch_changes_specs = (fields_spec,options)
    #~ else:
        #~ raise NotImplementedError()


def get_master(obj):
    
    #~ cs = WATCH_SPECS.get(self.watched.__class__)
    cs = obj._change_watcher_spec
    
    #~ print 20120921, cs
    if cs is None:
        return
    return cs.get_master(obj)
        
    #~ if cs.master_key is None:
        #~ return obj
    # master = cs.master_key.value_from_object(self.watched)
    # master = getattr(self.watched,cs.master_key)
    #~ return getattr(obj,cs.master_key.name)

def log_change(type,request,master,obj,msg=''):
    from lino.models import Change
    Change(
        type=type,
        time=datetime.datetime.now(),
        user=request.user,
        #~ summary=self.watched._change_summary,
        master=master,
        object=obj,
        diff=msg).save()


class Watcher(object):
    def __init__(self,watched):
        self.original_state = dict(watched.__dict__)
        self.watched = watched
        #~ self.is_new = is_new
        #~ self.request
        
    def is_dirty(self):
        #~ if self.is_new: 
            #~ return True
        for k,v in self.original_state.iteritems():
            if v != self.watched.__dict__.get(k, NOT_GIVEN):
                return True
        return False
        
    def log_diff(self,request):
        master = get_master(self.watched)
        if master is None:
            return
            
        #~ ignored_fields = self.watched._watch_changes_specs
        #~ ignored_fields            
        
        #~ watched_fields, options = cs
        
        #~ if self.is_new:
            #~ msg = u"%s created." % obj2str(self.watched,True)
            #~ msg = u"%s created by %s." % (obj2str(self.watched),request.user)
        #~ else:
        cs = self.watched._change_watcher_spec
        changes = []
        for k,old in self.original_state.iteritems():
            #~ if watched_fields is None or k in watched_fields:
            if not k in cs.ignored_fields:
                new = self.watched.__dict__.get(k, NOT_GIVEN)
                if old != new:
                    changes.append("%s : %s --> %s" % (k,obj2str(old),obj2str(new)))
        
        if len(changes) == 0:
            msg = '(no changes)'
        elif len(changes) == 1:
            msg = changes[0]
        else:
            msg = '- ' + ('\n- '.join(changes))
            
        log_change(ChangeTypes.update,request,master,self.watched,msg)
        
        
        

def log_delete(request,obj):
    """
    Calls :func:`log_change` with `ChangeTypes.delete`.
    
    Note that you must call this before actually deleting the object,
    otherwise mysql (not sqlite) says 
    ERROR: (1048, "Column 'object_id' cannot be null")
    """
    master = get_master(obj)
    if master is None:
        return
    log_change(ChangeTypes.delete,request,master,obj)
    
def log_create(request,obj):
    """
    To be called when a new instance has actually been created and saved.
    """    
    master = get_master(obj)
    if master is None:
        return
    log_change(ChangeTypes.create,request,master,obj,obj2str(obj,True))

def log_add_child(request,obj,child_model):
    master = get_master(obj)
    if master is None:
        return
    log_change(ChangeTypes.add_child,request,master,obj,full_model_name(child_model))

def log_remove_child(request,obj,child_model):
    master = get_master(obj)
    if master is None:
        return
    log_change(ChangeTypes.remove_child,request,master,obj,full_model_name(child_model))



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import decimal
from dateutil.relativedelta import relativedelta
ONE_DAY = relativedelta(days=1)

from django.db import models
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext as _


from lino.utils import i2d
from lino.utils.instantiator import Instantiator
from lino.core.modeltools import resolve_model
from lino.utils.babel import babel_values

from lino.core.perms import UserProfiles


from lino.modlib.debts.models import AccountType

def n2dec(v):
    return decimal.Decimal("%.2d" % v)
   
def objects():
    group = Instantiator('debts.AccountGroup').build
    g = group(account_type=AccountType.income,**babel_values('name',
          de=u"Monatliche Einkünfte",
          fr=u"Revenus mensuels",
          en=u"Monthly incomes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(required_for_person=True,**babel_values('name',
          de=u"Gehälter",
          fr=u"Salaires",
          en=u"Salaries"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Renten",
          fr=u"Pension",
          en=u"Pension"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Integrationszulage",
          fr=u"Allocation d'intégration",
          en=u"Integration aid"
          ))

    g = group(account_type=AccountType.income,**babel_values('name',
          de=u"Jährliche Einkünfte",
          fr=u"Revenus annuels",
          en=u"Yearly incomes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(required_for_person=True,**babel_values('name',
          de=u"Urlaubsgeld",
          fr=u"Congé payé",
          en=u"Paid holiday"
          ))
    yield account(required_for_person=True,**babel_values('name',
          de=u"Jahresendzulage",
          fr=u"Prime de fin d'année",
          en=u"Year-end prime"
          ))

    g = group(account_type=AccountType.expense,**babel_values('name',
          de=u"Monatliche Ausgaben",
          fr=u"Dépenses mensuelles",
          en=u"Monthly expenses"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(required_for_household=True,**babel_values('name',
          de=u"Miete",
          fr=u"Loyer",
          en=u"Rent"
          ))
    yield account(required_for_household=True,**babel_values('name',
          de=u"Strom",
          fr=u"Electricité",
          en=u"Electricity"
          ))


    g = group(account_type=AccountType.expense,**babel_values('name',
          de=u"Steuern",
          fr=u"Taxes",
          en=u"Taxes"
          ))
    yield g
    account = Instantiator('debts.Account',group=g,periods=12).build
    yield account(required_for_household=True,**babel_values('name',
          de=u"Müllsteuer",
          fr=u"Taxe déchets",
          en=u"Waste tax"
          ))

    g = group(account_type=AccountType.asset,**babel_values('name',
          de=u"Aktiva, Vermögen, Kapital",
          fr=u"Actifs",
          en=u"Assets"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(**babel_values('name',
          de=u"Vermögen",
          fr=u"Propriété",
          en=u"Assets"
          ))
    account = Instantiator('debts.Account',group=g).build
    yield account(**babel_values('name',
          de=u"Haus",
          fr=u"Maison",
          en=u"House"
          ))
    yield account(**babel_values('name',
          de=u"Auto",
          fr=u"Voiture",
          en=u"Car"
          ))
    
    
    g = group(account_type=AccountType.liability,**babel_values('name',
          de=u"Guthaben, Schulden, Verbindlichkeit",
          fr=u"Créances et dettes",
          en=u"Liabilities"
          ))
    yield g
    account = Instantiator('debts.Account',group=g).build
    yield account(**babel_values('name',
          de=u"Kredite",
          fr=u"Crédits",
          en=u"Loans"
          ))
    yield account(**babel_values('name',
          de=u"Schulden",
          fr=u"Emprunts",
          en=u"Debts"
          ))
    yield account(**babel_values('name',
          de=u"Gerichtsvollzieher",
          fr=u"Juge",
          en=u"Judge"
          ))
    yield account(**babel_values('name',
          de=u"Zahlungsrückstände",
          fr=u"Factures à payer",
          en=u"Invoices to pay"
          ))

    #~ budget = Instantiator('debts.Budget').build
    #~ from lino.modlib.users.models import User
    #~ root = User.objects.get(username='root')
    
    User = resolve_model('users.User')
    kerstin = User(username="kerstin",
        first_name="Kerstin",last_name=u"Küpper",
        profile=UserProfiles.kerstin)
    yield kerstin
    
    
    Household = resolve_model('households.Household')
    Budget = resolve_model('debts.Budget')
    Actor = resolve_model('debts.Actor')
    for hh in Household.objects.all():
        #~ sub_budgets = []
        for p in hh.member_set.all():
            yield Budget(partner_id=p.person.id,user=kerstin)
            #~ sub_budgets.append(b)
            #~ yield b
        yield Budget(partner_id=hh.id,user=kerstin)
        #~ yield b
        #~ for sb in sub_budgets:
            #~ yield Actor(budget=b,sub_budget=sb)
        
    Budget = resolve_model('debts.Budget')
    #~ Debt = resolve_model('debts.Debt')
    Entry = resolve_model('debts.Entry')
    Account = resolve_model('debts.Account')
    Company = resolve_model('contacts.Company')
    AMOUNTS = Cycler([i*5.24 for i in range(10)])
    PARTNERS = Cycler(Company.objects.all())
    ACCOUNTS = Cycler(Account.objects.filter(type=AccountType.liability))
    for b in Budget.objects.all():
        #~ n = min(3,b.actor_set.count())
        for e in b.entry_set.all():
            #~ for i in range(n):
            e.amount = n2dec(AMOUNTS.pop())
            e.save()
        for i in range(3):
            a = int(AMOUNTS.pop()*5)
            yield Entry(budget=b,
                account=ACCOUNTS.pop(),
                partner=PARTNERS.pop(),amount=a,
                monthly_rate=n2dec(a/20))
    
    


# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.


import os
import lino

from lino.apps.std.settings import *

class Lino(Lino):
    source_dir = os.path.dirname(__file__)
    title = "Lino/MinimalApp 2"
    #~ help_url = "http://lino.saffre-rumma.net/az/index.html"
    #~ migration_module = 'lino.apps.az.migrate'
    
    #~ project_model = 'contacts.Person'
    #~ project_model = 'contacts.Person'
    project_model = 'projects.Project'
    user_model = "users.User"
    
    #~ languages = ('de', 'fr')
    languages = ['en','et']
    
    #~ index_view_action = "dsbe.Home"
    
    #~ remote_user_header = "REMOTE_USER"
    #~ remote_user_header = None
    
    def get_app_source_file(self):  return __file__
        
    #~ def setup_quicklinks(self,ui,user,tb):
        #~ tb.add_action(self.modules.contacts.Persons.detail_action)
        #~ tb.add_action(self.modules.contacts.Companies.detail_action)

    def get_installed_apps(self):
        for a in super(Lino,self).get_installed_apps():
            yield a
        yield 'lino.modlib.users'
        yield 'lino.modlib.countries'
        yield 'lino.modlib.contacts'
        yield 'lino.modlib.projects'
        yield 'lino.modlib.uploads'
        yield 'lino.modlib.cal'
        yield 'lino.modlib.outbox'
        yield 'lino.modlib.pages'
        yield 'lino.apps.min2'

      
LINO = Lino(__file__,globals())



unused_INSTALLED_APPS = (
  #~ 'django.contrib.auth',
  'django.contrib.contenttypes',
  #~ 'django.contrib.sessions',
  #~ 'django.contrib.sites',
  #~ 'django.contrib.markup',
  #~ 'lino.modlib.system',
  'lino',
  'lino.modlib.users',
  'lino.modlib.countries',
  #~ 'lino.modlib.documents',
  #~ 'lino.modlib.properties',
  'lino.modlib.contacts',
  'lino.modlib.projects',
  #~ 'lino.apps.modlib.families',
  #~ 'lino.az.notes',
  #~ 'lino.apps.az.school',
  #~ 'lino.modlib.links',
  'lino.modlib.uploads',
  #~ 'lino.modlib.thirds',
  'lino.modlib.cal',
  #~ 'lino.modlib.mails',
  'lino.modlib.outbox',
  #~ 'lino.modlib.jobs',
  #~ 'lino.modlib.isip',
  #~ 'lino.modlib.bcss',
  #~ 'lino.modlib.newcomers',
  'lino.apps.min2',
)





## Copyright 2008-2010 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.


from django.db import models

from django.utils.safestring import mark_safe

from lino import dd
#~ from lino import reports
#~ from lino import layouts 
#~ from lino.utils import perms

#from lino.modlib.countries import models as countries 
countries = dd.resolve_app('countries')

class Place(dd.Model):
    name = models.CharField(max_length=200)
    country = models.ForeignKey(countries.Country)
    
    def __unicode__(self):
        return self.name
        

class PersonManager(models.Manager):
    def get_by_name(self,s):
        #print "get_by_name(%r)" % s
        try:
            return self.get(nickname__exact=s)
        except self.model.DoesNotExist,e:
            pass
        a = s.split()
        #~ if len(a) == 1: 
            #~ return self.get(nickname__exact=s)
        last_name = a.pop()
        for p in self.filter(last_name__exact=last_name):
            i = 0
            ok = True
            for fn in p.first_name.split():
                if a[i].endswith("."):
                    if not fn.startswith(a[i][:-1]):
                        ok = False
                elif i < len(a):
                    if a[i] != fn:
                        ok = False
                i += 1
            if ok:
                return p
        raise self.model.DoesNotExist("get_by_name(%r) failed" % s)
        
        
    
class Person(dd.Model):
    objects = PersonManager()
    first_name = models.CharField(max_length=100,blank=True)
    last_name = models.CharField(max_length=100,blank=True)
    #born = models.DateTimeField(blank=True,null=True)
    born = models.IntegerField(blank=True,null=True)
    died = models.IntegerField(blank=True,null=True)
    nickname = models.CharField(max_length=200,blank=True)
    name_prefix = models.CharField(max_length=10,blank=True)
    
    def __unicode__(self):
        #s = self.first_name + ' ' + self.last_name
        s = self.get_full_name()
        if self.born and self.died:
            s += " (%d-%d)" % (self.born,self.died)
        return s
        
    #~ def save(self, *args, **kwargs):
        #~ if len(self.nickname) == 0:
            #~ self.nickname = self.make_nickname()
        #~ super(Person,self).save(*args,**kwargs)
        
    def get_full_name(self):
        if len(self.nickname) > 0:
            return self.nickname
        l = [s for s in (self.first_name,
            self.name_prefix,
            self.last_name) if len(s)]
        return " ".join(l)
        #~ if self.first_name:
            #~ return self.first_name + " " + self.last_name
        #~ return self.last_name
    full_name = property(get_full_name)
        

class Author(Person):
    objects = PersonManager()

class Singer(Person):
    voice = models.CharField(max_length=10,blank=True,null=True)
    def __unicode__(self):
        s = self.first_name
        if self.voice is not None:
            s += " (%s)" % self.voice
        return s
    
class Choir(dd.Model):
    name = models.CharField(max_length=200)
    place = models.ForeignKey(Place)
    singers = models.ManyToManyField(Singer)
    
    def __unicode__(self):
        return self.name
        

class Song(dd.Model):
    title = models.CharField(max_length=200)
    #year_published = models.IntegerField(blank=True,null=True)
    text = models.TextField(blank=True)
    language = models.ForeignKey(countries.Language)
    origin = models.CharField(max_length=200)
    remarks = models.TextField(blank=True)
 
    composed_year = models.IntegerField(blank=True,null=True)
    text_year = models.IntegerField(blank=True,null=True)
    composed_by = models.ManyToManyField(
        Author,
        related_name="songs_composed")
    written_by = models.ManyToManyField(
        Author,
        related_name="songs_written")
    text_by = models.ManyToManyField(
        Author,
        related_name="texts_written")
    arranged_by = models.ManyToManyField(
        Author,
        related_name="songs_arranged")
    translated_by = models.ManyToManyField(
        Author,
        related_name="songs_translated")
    voices = models.CharField(max_length=200)
    bible_ref = models.CharField(max_length=200)
    based_on = models.ForeignKey("Song",blank=True,null=True)
    copyright = models.CharField(max_length=200)
        
    def __unicode__(self):
        return self.title + " (%d)" % self.id
        
class Event(dd.Model):
    date = models.DateField('date',blank=True,null=True)
    choir = models.ForeignKey(Choir)
    singers = models.ManyToManyField(Singer)
    #songs = models.ManyToManyField(Song,through="SongEvent")
    remark = models.TextField(blank=True)
    
    def add_song(self,song,**kw):
        if not kw.has_key('seq'):
            kw['seq'] = self.songevent_set.count() + 1
        if not isinstance(song,Song):
            song = Song.objects.get(pk=song)
        se = SongEvent(event=self,song=song,**kw)
        se.save()
        return se

    def add_songs(self,*songs,**kw):
        for song in songs:
            self.add_song(song,**kw)
            
    def __unicode__(self):
        s = str(self.date)
        if self.remark:
            s += " (%s)" % self.remark
        return s
        
class Rehearsal(Event):
    pass
    
class Performance(Event):
    place = models.ForeignKey(Place)
    
class SongEvent(dd.Model):
    event = models.ForeignKey(Event) #,related_name="songs")
    seq = models.IntegerField(default=0)
    song = models.ForeignKey(Song,related_name="events")
    remark = models.TextField(blank=True)
    
    def add_link(self,**kw):
        n = self.link_set.create(**kw)
        #n = Link(songevent=self,**kw)
        n.save()
        return n
        
class Link(dd.Model):
    url = models.URLField()
    text = models.CharField(max_length=200)
    #date = models.DateField('date',blank=True,null=True)
    songevent = models.ForeignKey(SongEvent)
    #song = models.ForeignKey(Song)
    #event = models.ForeignKey(Event,blank=True,null=True)

    def __unicode__(self):
        if self.text:
            return self.text
        return dd.Model.__unicode__(self)
        #if self.text:
        #    return HREF(self.url,self.text)
        #return HREF(self.url,self.url)
        
    def get_instance_url(self):
        # see sites.LinoSite.get_instance_url()
        if self.url:
            return self.url
        

#~ class Work(dd.Model):
    #~ class Admin:
        #~ pass
    #~ composer = models.ForeignKey(
        #~ Person,
        #~ related_name="songs_composed")
    #~ textauthor = models.ForeignKey(
        #~ Person,
        #~ related_name="texts_written")
    #~ song = models.ForeignKey(
        #~ Song,
        #~ related_name="composed_by")
    #~ remark = models.TextField(blank=True,null=True)
    
    
class Translation(dd.Model):
    class Admin:
        pass
    translator = models.ForeignKey(
        Person,
        related_name="translator",
        null=True)
    original = models.ForeignKey(
        Song,
        related_name="original",
        null=True)
    target = models.ForeignKey(
        Song,
        related_name="target",
        null=True)
    

class Collection(dd.Model):
    title = models.CharField(max_length=200)
    intro = models.TextField(blank=True,null=True)
    #publisher = models.TextField(blank=True,null=True)
    year = models.IntegerField(blank=True,null=True)
    author = models.ForeignKey(Author,blank=True,null=True)
    place = models.ForeignKey(Place,blank=True,null=True)
    songs = models.ManyToManyField(Song)
    
    def __unicode__(self):
        return self.title
    

# reports


class SongDetail(dd.FormLayout):
    datalink = 'songs.Song'
    main = """
    id title language voices
    origin based_on bible_ref
    composed_by text_by written_by
    arranged_by
    translated_by
    events
    """
    def inlines(self):
        return dict(events=EventsBySong())

class PerformanceDetail(dd.FormLayout):
    datalink = 'songs.Performance'
  
    main = """
    id date place remark 
    choir
    songs
    singers
    """
    def inlines(self):
        return dict(songs=SongsByEvent())
    
class Rehearsals(dd.Table):
    model = Rehearsal
    #page_layout_class = RehearsalDetail
    column_names = "date remark songs:20 singers:10"
    order_by = "date"

class Performances(dd.Table):
    model = Performance
    column_names = "date place remark songs:20 singers:10"
    order_by = "date"
    #~ page_layouts = (PerformanceDetail,)
    

class Collections(dd.Table):
    model = Collection
    column_names = "title place year author id"

class Places(dd.Table):
    model = Place
    column_names = "name country id"


class Singers(dd.Table):
    model = Singer
    #page_layout_class = RehearsalDetail
    column_names = "id first_name last_name voice"
    order_by = "last_name"

class Authors(dd.Table):
    model = Author
    #queryset = Author.objects.filter(singer__exact=None)
    #page_layout_class = RehearsalDetail
    column_names = "id get_full_name born died first_name last_name"
    order_by = "last_name"

    
class EventsBySong(dd.Table):
    model = SongEvent
    #master = Song
    master_key = 'song'
    column_names = "event remark link_set"
    
class SongsByEvent(dd.Table):
    model = SongEvent
    #master = Event
    master_key = 'event'
    column_names = "seq song link_set remark"

class Songs(dd.Table):
    model = Song
    #~ page_layouts = (SongDetail,)
    column_names = "id title language voices composed_by text_by written_by"

class SongEvents(dd.Table):
    model = SongEvent

class SongsByEvent(dd.Table):
    model = SongEvent
    #master = Event
    master_key = 'event'
    column_names = "seq song remark link_set"


class Links(dd.Table):
    model = Link

class Choirs(dd.Table):
    model = Choir


#~ def lino_setup(lino):
    #~ m = lino.add_menu("songs","~Songs",can_view=perms.is_authenticated)
    #~ m.add_action(Rehearsals())
    #~ m.add_action(Performances())
    #~ m.add_action(Singers())
    #~ m.add_action(Authors())
    #~ m.add_action(Songs())
    #~ m.add_action(Collections())
    #~ m.add_action(Places())
    #~ m.add_action(Links())
    #~ m.add_action(SongEvents())



# -*- coding: utf-8 -*-
## Copyright 2008-2009 Luc Saffre.
## This file is part of the Lino project. 

## Lino is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## Lino is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
## or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
## License for more details.

## You should have received a copy of the GNU General Public License
## along with Lino; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import os
import codecs
from models import Unit, Entry, Units
#from menu import Units
from django.core import serializers
from django.test import TestCase
from django import forms
from lino.utils.validatingmodel import ModelValidationError

# convert a django.forms.util.ErrorDict object to a str
#~ def errordict2str(errordict):
    #~ d={}
    #~ for fieldname,errorlist in errordict.items():
        #~ d[fieldname] = list(*errorlist)
    #~ return d


class TestCase(TestCase):
    def test01(self):
        
        u1 = Unit(title="First Chapter")
        u1.save()
        self.assertEqual(u1.seq,1)
        u = Unit(title="First Section",parent=u1)
        u.save()
        self.assertEqual(u.seq,1)
        u = Unit(title="Second Section",parent=u1)
        u.save()
        self.assertEqual(u.seq,2)
        
        self.assertEqual(u.pk,3)
        try:
            #print "set parent to self"
            u.parent=u
            u.save()
            print "save() : done"
        except ModelValidationError,e:
            #s="\n".join([e.as_text() for m in e.messages])
            self.assertEqual(str(e),"Parent cannot be self")
        else:
            self.fail("Expected ModelValidationError")
            
        
            

class PkkTestCase(TestCase):
    fixtures=[ 'demo' ]
    def setUp(self):
        for u in Unit.objects.all():
            u.save()
        
    def test01(self):
        unit=Unit.objects.get(pk=5)
        # unit.save()
        entries=unit.entry_set.all()
        s="\n".join([e.word1 for e in entries])
        #print s
        self.assertEquals(s.split(), u"""
        grand, grande
        gentil, gentille
        petit, petite
        méchant, méchante
        sale, sale
        propre, propre
        est-ce que
        oui
        non """.split())
        
    #~ def test02(self):
        #~ entries=Entry.objects.all()
        #~ self.assertEquals(len(entries),192)
        #~ entries=Entry.objects.filter(word1__startswith="p")
        #~ self.assertEquals(len(entries),20)
        #~ s="\n".join([e.word1 for e in entries])
        #~ print s
        #~ #print "test00",entries
        #~ self.assertEquals(s,"""\
#~ père
        #~ """)

    def test03(self):
        # total number of rows in each table:
        units=Unit.objects.all()
        self.assertEqual(len(units),6)
        self.assertEqual(len(Entry.objects.all()),9)
        
        pkk=unit=Unit.objects.get(pk=1)
        self.assertEqual(len(pkk.children.all()),1)
        
        #
        # prettyprint()
        #
        s=pkk.prettyprint()
        #print s
        self.assertEqual(s,u"""\
1. Prantsuse keele kurs algajatele
  1.1. Esimene tund
    1.1.1. Sissejuhatus
      1.1.2. Olema
      1.1.3. Esimesed laused
      1.1.4. Mees või naine?""")
      
        #
        #
        #
        
        entries=Entry.objects.filter(word1__startswith="p")
        self.assertEquals(len(entries),2)
        s="\n".join([e.word1 for e in entries])
        #print s
        self.assertEquals(s.split(),u"""
petit, petite
propre, propre
        """.split())
        
        #
        # The first Report
        #
        
        rpt=Units()
        s=rpt.as_text(column_widths=dict(id=3,title=30,parent=20,seq=3))
        #print "\n"+s
        self.assertEquals(s.split(),u"""
Units
=====
ID |title                         |name     |parent              |seq|format
---+------------------------------+---------+--------------------+---+---------
1  |Prantsuse keele kurs          |pkk      |                    |1  |R
   |algajatele                    |         |                    |   |
2  |Esimene tund                  |u1       |1. Prantsuse keele  |1  |R
   |                              |         |kurs algajatele     |   |
3  |Sissejuhatus                  |         |1.1. Esimene tund   |1  |R
4  |Olema                         |         |1.1. Esimene tund   |2  |R
5  |Esimesed laused               |         |1.1. Esimene tund   |3  |R
6  |Mees või naine?               |         |1.1. Esimene tund   |4  |R        
""".split(),"Units().as_text() has changed in demo")
        
        

        
     
        
# Run these tests using "python manage.py test".
# see http://docs.djangoproject.com/en/dev/topics/testing/#topics-testing



"""

>>> luc = contacts.Contact(first_name="Luc",last_name="Saffre")
>>> luc.save()
>>> luc
<Contact: Luc Saffre>

>>> cust = Customer(contact=luc)
>>> cust.save()
>>> cust.last_name
Saffre

# the following fails because of ticket #7626
# http://code.djangoproject.com/ticket/7623

"""

from django.db import models
from lino.apps.sandbox.contacts import models as contacts

# extends Contact

class Customer(contacts.Contact):
    contact = models.OneToOneField(contacts.Contact,parent_link=True)
    payment_term = models.IntegerField(blank=True,null=True)
    



## Copyright 2009-2011 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import os

from django.utils.translation import ugettext as _
from django.db import models
from django.conf import settings

import lino
from lino.core import actions
#~ from lino import forms
from lino.ui import base
from lino.core import actors
#~ from lino.utils import choosers
from lino.utils import jsgen
#~ from lino.utils import build_url
from lino.utils.jsgen import py2js, js_code, id2js
from . import ext_elems
from lino.ui import requests as ext_requests

#~ from lino.ui.extjs import ext_viewport

#~ from lino.modlib.properties import models as properties

#~ WC_TYPE_GRID = 'grid'
#~ USE_FF_CONSOLE = True

class ActionRenderer(object):
    def __init__(self,ui,action):
        assert isinstance(action,actions.Action), "%r is not an Action" % action
        self.action = action
        self.ui = ui
        
    def update_config(self,wc):
        pass
        
    def js_render(self):
        yield "Lino.%s = function(caller) { " % self.action
        yield "    return new Lino.%s(caller,%s);}" % (
            self.__class__.__name__,py2js(self.config))

class WindowWrapper(ActionRenderer):
   
    def __init__(self,action,ui,lh,main,**kw):
        ActionRenderer.__init__(self,ui,action)
        self.main = main
        self.lh = lh # may be None
        #~ self.bbar_buttons = []
        self.config = self.get_config()
        
    def __str__(self):
        return self.ext_name + "(" + self.__class__.__name__ + ")"
        
    def get_config(self,**d):
        #~ d.update(permalink_name=str(self.action))
        #~ d.update(action_name=self.action.name)
        return d
        

#~ class MainComponent(Component): 
    #~ def js_value(self):
        #~ assert not self.ext_options()
        value = self.ext_options(**self.value)
        #~ yield self.value_template % 'params'
  


class MasterWrapper(WindowWrapper):
  
    def __init__(self,rh,action,lh,**kw):
        WindowWrapper.__init__(self,action,lh.rh.ui,lh,lh._main,**kw)
        
    def js_render(self):
        yield "Lino.%s = function(caller,params) { " % self.action
        #~ yield "function(caller,params) { "
        #~ yield "  Ext.getCmp('main_area').el.setStyle({cursor:'wait'});"
        #~ yield "Lino.notify();"
        if False and settings.USE_FIREBUG:
            yield "  console.time('%s');" % self.action
            #~ yield "  console.log('ext_windows',20100930,params);"
        yield "  var ww = new Lino.%s(caller,%s);" % (
            self.__class__.__name__,py2js(self.config))
            
        
        """HACK : override default jsgen behaviour. 
        The main component has no options, but must take those 
        passed to this function as `params`.
        """
        if self.main.value:
            raise Exception("Oops, value is %r" % self.main.value)
        v = self.main
        yield "  params.containing_window = ww;"
        yield "  var %s = %s;" % (v.ext_name,(v.value_template % 'params'))
            
        
        #~ self.main.value = 'params' ## ugly HACK!
        
        #~ for ln in jsgen.declare_vars(self.main):
            #~ yield '  ' + ln
            
        yield "  ww.main_item = %s;" % self.main.as_ext()
            
        yield "  ww.show();"
        if False and settings.USE_FIREBUG:
            yield "  console.timeEnd('%s');" % self.action
        yield "};"
        
            
    
class GridWrapperMixin(WindowWrapper):
  
    #~ window_config_type = WC_TYPE_GRID
    
    def __init__(self,rh):
        self.rh = rh
    
    def get_config(self):
        d = super(GridWrapperMixin,self).get_config()
        #~ d.update(content_type=self.rh.content_type)
        #~ d.update(title=unicode(self.rh.get_title(None)))
        #~ 20101022 d.update(main_panel=self.lh._main)
        return d
        
    def u_update_config(self,wc):
        self.lh._main.update_config(wc)
        

        

class CalendarWrapper(MasterWrapper):
    def __init__(self,rh,action,**kw):
        main = ext_elems.Calendar()
        WindowWrapper.__init__(self,action,rh.ui,None,main,**kw)
    
    
class GridMasterWrapper(GridWrapperMixin,MasterWrapper):
  
    def __init__(self,rh,action,**kw):
        self.action = action
        GridWrapperMixin.__init__(self,rh)
        MasterWrapper.__init__(self,rh,action,rh.list_layout,**kw)


class DetailWrapper(MasterWrapper):
  
    def __init__(self,rh,action,**kw):
        self.rh = rh
        main = ext_elems.FormPanel(rh,action)
        WindowWrapper.__init__(self,action,rh.ui,None,main,**kw)
        
        
    def get_config(self):
        d = MasterWrapper.get_config(self)
        # url = self.ui.build_url('api',self.action.actor.app_label,self.action.actor._actor_name)
        # d.update(content_type=self.rh.content_type)
        # d.update(active_fields=self.rh.report.active_fields) 
        # d.update(url_data=url) 
        # 20101022 d.update(main_panel=self.main)
        # d.update(name=self.action.name) # used by tinymce editor window
        # d.update(master_key=self.action.actor.master_key);
        return d
        
        
        
  
class InsertWrapper(DetailWrapper):
    pass

    #~ def get_config(self):
        #~ d = DetailWrapper.get_config(self)
        #~ # d.update(record_id=-99999);
        #~ return d
        


def key_handler(key,h):
    return dict(handler=h,key=key.keycode,ctrl=key.ctrl,alt=key.alt,shift=key.shift)





## Copyright 2011-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

r"""
A shortcut to classes and methods needed 
when defining your database structure in a `models` module.

The name "dd" stands for "Data Definition".

A small wrapper around Django's `Model` class 
which adds some Lino specific features:

- :class:`Model <lino.core.model.Model>`

Tables:

- :class:`Table <lino.core.dbtables.Table>`
- :class:`VirtualTable <lino.core.tables.VirtualTable>`
- :class:`Frame <lino.core.frames.Frame>`

Extended Fields:

- :class:`NullCharField <lino.core.fields.NullCharField>`
- :class:`IncompleteDateField <lino.core.fields.IncompleteDateField>`
- :class:`PasswordField <lino.core.fields.PasswordField>`
- :class:`MonthField <lino.core.fields.MonthField>`
- :class:`QuantityField <lino.core.fields.QuantityField>`
- :class:`PriceField<lino.core.fields.PriceField>`
- :class:`GenericForeignKey <lino.core.fields.GenericForeignKey>`
- :class:`GenericForeignKeyIdField <lino.core.fields.GenericForeignKeyIdField>`
- :class:`RecurrenceField <lino.core.fields.RecurrenceField>`
- :class:`DummyField <lino.core.fields.DummyField>`
- :func:`ForeignKey <lino.core.fields.ForeignKey>`

Babel fields:

- :class:`BabelCharField <lino.utils.babel.BabelCharField>`
- :class:`BabelTextField <lino.utils.babel.BabelTextField>`

Virtual Fields:

- :class:`Constant <lino.core.fields.Constant>` and 
  :class:`@constant <lino.core.fields.constant>`
- :class:`DisplayField <lino.core.fields.DisplayField>` and 
  :class:`@displayfield <lino.core.fields.displayfield>`
- :class:`VirtualField <lino.core.fields.VirtualField>` and
  :class:`@virtualfield <lino.core.fields.virtualfield>`
- :class:`HtmlBox <lino.core.fields.HtmlBox>`

Layouts:

- :class:`FormLayout <lino.core.layouts.FormLayout>`
- :class:`Panel <lino.core.layouts.Panel>`
  
Utilities:

- :func:`fields_list <lino.core.fields.fields_list>`
- :func:`resolve_field <lino.core.modeltools.resolve_field>`
- :func:`resolve_model <lino.core.modeltools.resolve_model>`
- :func:`resolve_app <lino.core.modeltools.resolve_app>` 
- :func:`chooser <lino.utils.choosers.chooser>` 
- :func:`add_user_group` 
- :func:`inject_quick_add_buttons` 
- :func:`update_field` 
- :func:`inject_field` 

Actions:

- :class:`RowAction <lino.core.actions.RowAction >`
- :class:`ChangeStateAction <lino.core.changes.ChangeStateAction>`
- :class:`NotifyingAction <lino.core.actions.NotifyingAction>`
- :class:`AuthorRowAction <lino.mixins.AuthorRowAction>`

Miscellaneous:

- :class:`ChoiceList <lino.core.choicelists.ChoiceList>`
- :class:`Workflow <lino.core.workflows.Workflow>`


- :class:`UserProfiles <lino.utils.auth.UserProfiles>`
- :class:`UserGroups <lino.utils.auth.UserGroups>`
- :class:`UserLevels <lino.utils.auth.UserLevels>`


"""

import logging
logger = logging.getLogger(__name__)


from lino.core.tables import VirtualTable

from lino.core.modeltools import resolve_model, resolve_app, resolve_field, get_field, UnresolvedModel
from lino.core.modeltools import full_model_name
from lino.core.modeltools import models_by_base

from lino.core.model import Model

#~ from lino.core.table import fields_list, inject_field
from lino.core.dbtables import has_fk
from lino.core.dbtables import Table
from django.db.models.fields import FieldDoesNotExist
from django.db import models
from django.conf import settings
#~ Model = models.Model
#~ from lino.core import table
#~ Table = table.Table

from lino.core.dbtables import summary, summary_row

from lino.core.frames import Frame
from lino.core.dialogs import Dialog

from lino.core.actions import action
#~ from lino.core.actions import Action
from lino.core.actions import RowAction
from lino.mixins import AuthorRowAction
#~ from lino.core.actions import ListAction
from lino.core.actions import GridEdit, ShowDetailAction
from lino.core.actions import InsertRow, DeleteSelected
from lino.core.actions import SubmitDetail, SubmitInsert
#~ from lino.core.actions import Calendar

from lino.core.choicelists import ChoiceList, Choice
from lino.core.workflows import Workflow, ChangeStateAction
#~ from lino.core.changes import ChangeStateAction
from lino.core.actions import NotifyingAction


from lino.core.fields import fields_list
from lino.core.fields import DummyField
from lino.core.fields import RecurrenceField
from lino.core.fields import GenericForeignKey
from lino.core.fields import GenericForeignKeyIdField
from lino.core.fields import IncompleteDateField
from lino.core.fields import DatePickerField
from lino.core.fields import NullCharField
from lino.core.fields import PasswordField
from lino.core.fields import MonthField
#~ from lino.core.fields import LinkedForeignKey
from lino.core.fields import QuantityField
from lino.core.fields import HtmlBox, PriceField, RichTextField

from lino.core.fields import DisplayField, displayfield
from lino.core.fields import VirtualField, virtualfield
from lino.core.fields import RequestField, requestfield
from lino.core.fields import Constant, constant
from lino.core.fields import ForeignKey

from lino.utils.babel import BabelCharField, BabelTextField
#~ from lino.core.fields import MethodField

from lino.utils.choosers import chooser

from lino.utils.auth import UserLevels, UserProfiles, UserGroups

#~ from lino.base.utils import UserLevels, UserGroups, UserProfiles

from lino.core.layouts import FormLayout, Panel
from lino.core.layouts import ParamsLayout
#~ from lino.core.layouts import ActionParamsLayout
#~ DetailLayout = InsertLayout = FormLayout

#~ from lino.core.layouts import DetailLayout, InsertLayout


class Module(object):
    pass
    
    
  
PENDING_INJECTS = dict()
PREPARED_MODELS = dict()

def fix_field_cache(model):
    """
    Remove duplicate entries in `_field_cache` to fix Django issue #10808
    """
    cache = []
    field_names = set()
    for f,m in model._meta._field_cache:
        if f.attname not in field_names:
            field_names.add(f.attname)
            cache.append( (f,m) )
    model._meta._field_cache = tuple(cache)
    model._meta._field_name_cache = [x for x, _ in cache]
    #~ logger.info("20130106 fixed field_cache %s (%s)",full_model_name(model),' '.join(field_names))


def on_class_prepared(signal,sender=None,**kw):
    """
    This is Lino's general `class_prepared` handler.
    It does two things:
    
    - Run pending calls to :func:`inject_field` and :func:`update_field`.
    
    - Apply a workaround for Django's ticket 10808.
      In a Diamond inheritance pattern, `_meta._field_cache` contains certain fields twice.    
      So we remove these duplicate fields from `_meta._field_cache`.
      (A better solution would be of course to not collect them.)
    """
    model = sender
    #~ return
    #~ if model is None:
        #~ return 
    k = model._meta.app_label + '.' + model.__name__
    PREPARED_MODELS[k] = model
    #~ logger.info("20120627 on_class_prepared %r = %r",k,model)
    todos = PENDING_INJECTS.pop(k,None)
    if todos is not None:
        for f in todos:
            f(model)
        #~ for k,v in injects.items():
            #~ model.add_to_class(k,v)

    """
    django.db.models.options
    """
    if hasattr(model._meta,'_field_cache'):
        fix_field_cache(model)
    #~ else:
        #~ logger.info("20120905 Could not fix Django issue #10808 for %s",model)


models.signals.class_prepared.connect(on_class_prepared)

def on_analyze_models(): # called from kernel.analyze_models()
    
    if PENDING_INJECTS:
        msg = ''
        for spec,funcs in PENDING_INJECTS.items():
            msg += spec + ': ' 
            #~ msg += '\n'.join([str(dir(func)) for func in funcs])
            #~ msg += '\n'.join([str(func.func_code.co_consts) for func in funcs])
            msg += str(funcs)
        raise Exception("Oops, there are pending injects: %s" % msg)
        #~ logger.warning("pending injects: %s", msg)

    """
    20130106
    now we loop a last time over each model and fill it's _meta._field_cache
    otherwise if some application module used inject_field() on a model which 
    has subclasses, then the new field would not be seen by subclasses
    """
    for model in models.get_models():
        model._meta._fill_fields_cache()
        fix_field_cache(model)

from lino.core.modeltools import is_installed_model_spec

def do_when_prepared(model_spec,todo):
    if model_spec is None:
        return # e.g. inject_field during autodoc when user_model is None
        
    if isinstance(model_spec,basestring):
        if not is_installed_model_spec(model_spec):
            return
        k = model_spec
        model = PREPARED_MODELS.get(k,None)
        if model is None: 
            injects = PENDING_INJECTS.setdefault(k,[])
            injects.append(todo)
            #~ d[name] = field
            #~ logger.info("20120627 Defer inject_field(%r,%r,%r)", model_spec,name,field)
            return
    else:
        model = model_spec
        #~ k = model_spec._meta.app_label + '.' + model_spec.__name__
    todo(model)


def inject_field(model_spec,name,field,doc=None):
    """
    Adds the given field to the given model.
    See also :doc:`/tickets/49`.
    
    Since `inject_field` is usually called at the global level 
    of `models modules`, it cannot know whether the given `model_spec` 
    has already been imported (and its class prepared) or not. 
    That's why it uses Django's `class_prepared` signal to maintain 
    its own list of models.
    """
    #~ logger.info("20130106 inject_field(%r,%s)",model_spec,name)
    if doc:
        field.__doc__ = doc
    def todo(model):
        model.add_to_class(name,field)
        #~ if hasattr(model._meta,'_field_cache'):
        model._meta._fill_fields_cache()
        #~ fix_field_cache(model)
        #~ for m in models_by_base(model):
            #~ if hasattr(m._meta,'_field_cache'):
                #~ m._meta._fill_fields_cache()
                #~ fix_field_cache(m)
        #~ else:
            #~ logger.info("20130106 no need to fix_field_cache after inject_field")

    return do_when_prepared(model_spec,todo)    
    
    

def update_field(model_spec,name,**kw):
    """
    Update some attribute of the specified existing field.
    For example 
    :class:`PersonMixin <lino.modlib.contacts.models.PersonMixin>` 
    defines a field `first_name` which may not be blank.
    If you inherit from 
    :class:`PersonMixin <lino.modlib.contacts.models.PersonMixin>`
    but want `first_name` to be optional::
    
      class MyPerson(contacts.PersonMixin):
        ...
      dd.update_field(MyPerson,'first_name',blank=True)
      
    Or you want to change the label of a field defined in an inherited mixin,
    as done in  :app:`lino.modlib.outbox.models`::
    
      dd.update_field(Mail,'user',verbose_name=_("Sender"))
    
    """
    def todo(model):
        try:
            fld = model._meta.get_field(name)
            #~ fld = model._meta.get_field_by_name(name)[0]
        except FieldDoesNotExist:
            logger.warning("Cannot update unresolved field %s.%s", model,name)
            return
        if fld.model != model:
            raise Exception('20120715 update_field(%s.%s) : %s' % (model,fld,fld.model))
            #~ logger.warning('20120715 update_field(%s.%s) : %s',model,fld,fld.model)
        for k,v in kw.items():
            setattr(fld,k,v)
    return do_when_prepared(model_spec,todo)    
        

def inject_quick_add_buttons(model,name,target):
    """
    Injects a virtual display field `name` into the specified `model`.
    This field will show up to three buttons
    `[New]` `[Show last]` `[Show all]`. 
    `target` is the table that will run these actions.
    It must be a slave of `model`.
    """
    def fn(self,rr):
        return rr.renderer.quick_add_buttons(
          rr.spawn(target,master_instance=self))
    inject_field(model,name,
        VirtualField(DisplayField(
            target.model._meta.verbose_name_plural),fn))
    
def add_user_group(name,label):
    """
    Add a user group to the :class:`UserGroups <lino.core.perms.UserGroups>` 
    choicelist. If a group with that name already exists, add `label` to the 
    existing group.
    """
    #~ logging.info("add_user_group(%s,%s)",name,label)
    #~ print "20120705 add_user_group(%s,%s)" % (name,unicode(label))
    g = UserGroups.items_dict.get(name)
    if g is None:
        UserGroups.add_item(name,label)
    else:
        if g.text != label:
            g.text += " & " + unicode(label)
    
def required(**kw):
    #~ if not kw.has_key('auth'):
        #~ kw.update(auth=True)
    kw.setdefault('auth',True)
    return kw

    


## Copyright 2009-2013 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""modeltools may not be used in models modules (but well during lino_site set up).
"""

import logging
logger = logging.getLogger(__name__)


import os
import sys
import datetime

from django.db import models
from django.conf import settings
from django.db.models import Q
from django.db.models.fields import FieldDoesNotExist
from django.utils.importlib import import_module
from django.db.models import loading


from django.core.validators import validate_email, ValidationError, URLValidator
validate_url = URLValidator()
def is_valid_url(s):
    try:
        validate_url(s)
        return True
    except ValidationError:
        return False
        
def is_valid_email(s):
    try:
        validate_email(s)
        return True
    except ValidationError:
        return False
        
        
        
def is_devserver():
    """
    Returns True if we are running a development server.
    
    Thanks to Aryeh Leib Taurog in 
    `How can I tell whether my Django application is running on development server or not?
    <http://stackoverflow.com/questions/1291755>`_
    
    Added the `len(sys.argv) > 1` test because in a 
    wsgi application the process is called without arguments.
    """
    return len(sys.argv) > 1 and sys.argv[1] == 'runserver'


def resolve_app(app_label,strict=False):
    """
    Return the `modules` module of the given `app_label` if it is installed. 
    Otherwise return either the :term:`dummy module` for `app_label` 
    if it exists, or `None`.
    If the optional second argument `strict` is `True`, then 
    
    This function is designed for use in models modules and available 
    through the shortcut ``dd.resolve_app``.
    
    For example, instead of writing::
    
        from lino.modlib.contacts import models as contacts
        
    it is recommended to write::
        
        contacts = dd.resolve_app('contacts')
        
    Because it makes your code usable 
    (1) in applications that don't have the 'contacts' module installed
    and
    (2) in applications who have another implementation of the contacts module.
    
    """
    #~ app_label = app_label
    for app_name in settings.INSTALLED_APPS:
        if app_name == app_label or app_name.endswith('.'+app_label):
            return import_module('.models', app_name)
    #~ return import_module('lino.modlib.%s.dummy' % app_label)
    try:
        return import_module('lino.modlib.%s.dummy' % app_label)
    except ImportError:
        if strict: 
            raise
            #~ raise Exception("strict resolve_app failed for app_label %r" % app_label)
        
    #~ if not emptyOK:
    #~ raise ImportError("No application labeled %r." % app_label)
#~ resolve_app = get_app

#~ def get_models_for(app_label):
    #~ a = models.get_app(app_label)
    
class UnresolvedModel:
    """
    This is the object returned by :func:`resolve_model` 
    if the specified model is not installed.
    
    We don't want resolve_model to raise an Exception because there are 
    cases of :doc:`data migration </topics/datamig>` where it would 
    disturb. 
    Asking for a non-installed model is not a sin, but trying to use it is.
    
    I didn't yet bother very much about finding a way to make the 
    model_spec appear in error messages such as
    ``AttributeError: UnresolvedModel instance has no attribute '_meta'``.
    Current workaround is to uncomment the ``print`` statement 
    below in such situations...
    
    """
    def __init__(self,model_spec,app_label):
        self.model_spec = model_spec
        self.app_label = app_label
        #~ print repr(self)
        
    def __repr__(self):
        return self.__class__.__name__ + '(%s,%s)' % (self.model_spec,self.app_label)
        
    #~ def __getattr__(self,name):
        #~ raise AttributeError("%s has no attribute %r" % (self,name))

def resolve_model(model_spec,app_label=None,strict=False,seed_cache=True):
    """
    See also django.db.models.fields.related.add_lazy_relation()
    """
    #~ models.get_apps() # trigger django.db.models.loading.cache._populate()
    if isinstance(model_spec,basestring):
        if '.' in model_spec:
            app_label, model_name = model_spec.split(".")
        else:
            model_name = model_spec
            
        #~ try:
            #~ app_label, model_name = model_spec.split(".")
        #~ except ValueError:
            #~ # If we can't split, assume a model in current app
            #~ #app_label = rpt.app_label
            #~ model_name = model_spec
            
        #~ model = models.get_model(app_label,model_name,seed_cache=False)
        model = models.get_model(app_label,model_name,seed_cache=seed_cache)
    else:
        model = model_spec
    if not isinstance(model,type) or not issubclass(model,models.Model):
        if strict:
            if isinstance(strict,basestring):
                raise Exception(strict % model_spec)
            raise Exception(
                "resolve_model(%r,app_label=%r) found %r (settings %s)" % (
                model_spec,app_label,model,settings.SETTINGS_MODULE))
        #~ logger.info("20120628 unresolved %r",model)
        return UnresolvedModel(model_spec,app_label)
    return model
    
def old_resolve_model(model_spec,app_label=None,strict=False):
    """
    doesn't work for contacts.Company because the model is defined somewhere else.
    """
    models.get_apps() # trigger django.db.models.loading.cache._populate()
    if isinstance(model_spec,basestring):
        if '.' in model_spec:
            app_label, model_name = model_spec.split(".")
        else:
            model_name = model_spec
        app = resolve_app(app_label)
        model = getattr(app,model_name,None)
    else:
        model = model_spec
    if not isinstance(model,type) or not issubclass(model,models.Model):
        if strict:
            raise Exception(
                "resolve_model(%r,app_label=%r) found %r (settings %s)" % (
                model_spec,app_label,model,settings.SETTINGS_MODULE))
        return UnresolvedModel(model_spec,app_label)
    return model
    
def get_field(model,name):
    '''Returns the field descriptor of the named field in the specified model.
    '''
    for vf in model._meta.virtual_fields:
        if vf.name == name:
            return vf
    fld, remote_model, direct, m2m = model._meta.get_field_by_name(name)
    # see blog/2011/0525
    #~ if remote_model is not None:
        #~ raise Exception("get_field(%r,%r) got a remote model ?!" % (model,name))
    return fld
  
class UnresolvedField(object):
    """
    Returned by :func:`resolve_field` if the specified field doesn't exist.
    This case happens when sphinx autodoc tries to import a module.
    See ticket :doc:`/tickets/4`.
    """
    def __init__(self,name):
        self.name = name
        self.verbose_name = "Unresolved Field " + name

def resolve_field(name,app_label=None):
    """
    Returns the field descriptor specified by the string `name` which 
    should be either `model.field` or `app_label.model.field`.
    """
    l = name.split('.')
    if len(l) == 3:
        app_label = l[0]
        del l[0]
    if len(l) == 2:
        #print "models.get_model(",app_label,l[0],False,")"
        #~ model = models.get_model(app_label,l[0],False)
        model = models.get_model(app_label,l[0])
        if model is None:
            raise FieldDoesNotExist("No model named '%s.%s'" % (app_label,l[0]))
        fld, remote_model, direct, m2m = model._meta.get_field_by_name(l[1])
        #~ try:
            #~ fld, remote_model, direct, m2m = model._meta.get_field_by_name(l[1])
        #~ except FieldDoesNotExist:
            #~ return UnresolvedField(name)
        assert remote_model is None or issubclass(model,remote_model), \
            "resolve_field(%r) : remote model is %r (expected None or base of %r)" % (name,remote_model,model)
        return fld
    raise FieldDoesNotExist(name)
    #~ return UnresolvedField(name)


#~ def requires_apps(self,*app_labels):
    #~ for app_label in app_labels:
        #~ get_app(app_label)
    
    
    
def full_model_name(model,sep='.'):
    """Returns the "full name" of the given model, e.g. "contacts.Person" etc.
    """
    return model._meta.app_label + sep + model._meta.object_name
    
    
    
def obj2unicode(i):
    """Returns a user-friendly unicode representation of a model instance."""
    return u'%s "%s"' % (i._meta.verbose_name,unicode(i))
    
def obj2str(i,force_detailed=False):
    """
    Returns a human-readable ascii string representation of a model instance, 
    even in some edge cases.
    """
    if not isinstance(i,models.Model): 
        if isinstance(i,long): return str(i) # AutoField is long on mysql, int on sqlite
        if isinstance(i,datetime.date): return i.isoformat()
        if isinstance(i,unicode): return repr(i)[1:]
        return repr(i)
    if i.pk is None:
        force_detailed = True
    if not force_detailed:
        if i.pk is None:
            return '(Unsaved %s instance)' % (i.__class__.__name__)
        try:
            return u"%s #%s (%s)" % (i.__class__.__name__,str(i.pk),repr(unicode(i)))
        except Exception,e:
        #~ except TypeError,e:
            return "Unprintable %s(pk=%r,error=%r" % (
              i.__class__.__name__,i.pk,e)
            #~ return unicode(e)
    #~ names = [fld.name for (fld,model) in i._meta.get_fields_with_model()]
    #~ s = ','.join(["%s=%r" % (n, getattr(i,n)) for n in names])
    pairs = []
    for (fld,model) in i._meta.get_fields_with_model():
        #~ if fld.name == 'language':
            #~ print 20120905, model, fld
        if isinstance(fld,models.ForeignKey):
            v = getattr(i,fld.attname) 
            #~ v = getattr(i,fld.name+"_id") 
            #~ if getattr(i,fld.name+"_id") is not None:
                #~ v = getattr(i,fld.name)
        else:
            v = getattr(i,fld.name)
        if v:
            pairs.append("%s=%s" % (fld.name,obj2str(v)))
    s = ','.join(pairs)
    #~ s = ','.join(["%s=%s" % (n, obj2str(getattr(i,n))) for n in names])
    #~ print i, i._meta.get_all_field_names()
    #~ s = ','.join(["%s=%r" % (n, getattr(i,n)) for n in i._meta.get_all_field_names()])
    return "%s(%s)" % (i.__class__.__name__,s)
    #~ return "%s(%s)" % (i.__class__,s)


def sorted_models_list():
    models_list = models.get_models() # trigger django.db.models.loading.cache._populate()
    def fn(a,b):
        return cmp(full_model_name(a),full_model_name(b))
    models_list.sort(fn)
    return models_list

def models_by_base(base):
    """
    Yields a list of installed models that are 
    subclass of the given base class.
    """
    for m in models.get_models():
        if issubclass(m,base):
            yield m
    
#~ models_by_abc = models_by_base

def makedirs_if_missing(dirname):
    """
    Make missing directories if they don't exist 
    and if :attr:`lino.Lino.make_missing_dirs` 
    is `True`.
    """
    #~ if not os.path.exists(dirname):
        #~ os.makedirs(dirname)
    if not os.path.isdir(dirname):
        if settings.LINO.make_missing_dirs:
            os.makedirs(dirname)
        else:
            raise Exception("Please create yourself directory %s" % dirname)
        
def range_filter(v,f1,f2):
    """
    Returns a Q object (to be added as a filter on a queryset)
    to inlude only instances where v is contained within the range between f1 and f2.
    `v` being a value and f1 and f2 being the names of fields of same data type as v.
    """
    #~ filter = Q(**{f2+'__isnull':False}) | Q(**{f1+'__isnull':False})
    q1 = Q(**{f1+'__isnull':True}) | Q(**{f1+'__lte':v})
    q2 = Q(**{f2+'__isnull':True}) | Q(**{f2+'__gte':v})
    return Q(q1,q2)
  




def app_labels():
    return [a.__name__.split('.')[-2] for a in loading.get_apps()]
        
#~ def get_slave(model,name):
    #~ """Return the named table, knowing that it is a 
    #~ slave of the specified `model`. 
    #~ If name has no app_label specified, use the model's app_label.
    #~ """
    #~ if not '.' in name:
        #~ name = model._meta.app_label + '.' + name
    #~ rpt = actors.get_actor(name)
    #~ if rpt is None: 
        #~ return None
    #~ return rpt

def get_model_report(model):
    if not hasattr(model,'_lino_default_table'):
        raise Exception("%r has no _lino_default_table" % model)
    return model._lino_default_table

    
def is_installed_model_spec(model_spec):
    app_label, model_name = model_spec.split(".")
    return settings.LINO.is_installed(app_label)
    



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
Does the same as the original Django dumpdata command,
but calls `lino.Lino.startup` first.
This is necessary if your application has its 
own :meth:`lino.Lino.setup_choicelists`.

"""

from django.conf import settings

from django.core.management.commands.dumpdata import Command as DumpdataCommand


#~ from lino.core.kernel import analyze_models

class Command(DumpdataCommand):
  
    def handle(self, *app_labels, **options):
        #~ analyze_models()
        settings.LINO.startup()
        
        options.update(format='py')
      
        return DumpdataCommand.handle(self, *app_labels, **options)
  



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
Set password "1234" for all users.

This is an additive fixture designed to work also on existing data.

  T:\data\luc\lino_local\dsbe>python manage.py loaddata demo2
  INFO Analyzing models...
  INFO Loading t:\hgwork\lino\lino\modlib\users\fixtures\demo2.py...
  Installed 8 object(s) from 1 fixture(s)



"""

from django.conf import settings

def objects():
    for u in settings.LINO.user_model.objects.exclude(profile=''):
        u.set_password('1234')
        yield u
                
